mkAnnotPkg <- function(filename, pkgName) {

  library(AnnBuilder)
  basefilename <- filename
  pkgpath <- .path.package("AnnBuilder")

  makeSrcInfo()
  srcObjs <- list()

  # source URLs (can change often)
  # entrez gene
  egUrl <- "ftp://ftp.ncbi.nlm.nih.gov/refseq/LocusLink/ARCHIVE"
  # unigene
  ugUrl <- "ftp://ftp.ncbi.nih.gov/repository/UniGene/Homo_Sapiens/Hs.data.gz"
  # goldenpath
  gpUrl <- "ftp://hgdownload.cse.ucsc.edu/goldenPath/hg17/database/"
  # kegg
  kgUrl <- "ftp://ftp.genome.ad.jp/pub/kegg/pathways"


  # classes for annotation data
  eg <- EG(srcUrl=egUrl, parser=file.path(pkgpath, "scripts", "gbLLParser"),
	baseFile=basefilename, accession="LL_tmpl.gz", built="N/A",
	fromWeb=TRUE)
  ug <- UG(srcUrl=ugUrl, parser=file.path(pkgpath, "scripts", "gbUGParser"),
	baseFile=basefilename, organism="Homo sapiens", built="N/A",
	fromWeb=TRUE)
  gp <- GP(srcUrl=gpUrl, organism="Homo sapiens", fromWeb=T)
  kg <- KEGG(srcUrl=kgUrl, organism="Homo sapiens", fromWeb=T)

  # source objects for gathering entrez gene ids and unigene with help of genebank ids
  srcObjs[["eg"]] <- eg
  srcObjs[["ug"]] <- ug

  # map entrez gene and uni gene ids to affy ids
  llMapping <- parseData(eg, eg@accession)
  colnames(llMapping) <- c("PROBE", "EG")
  ugMapping <- parseData(ug)
  colnames(ugMapping) <- c("PROBE", "UG")

  # merge affy, entrez and uni gene ids together to get a unique affy id -> entrez id map
  merged <- matrix(scan(basefilename, what="", sep="\t", quote="", quiet=TRUE), ncol=2, byrow=TRUE)
  colnames(merged) <- c("PROBE", "ACC")

  if (dim(llMapping)[1]==0 & dim(ugMapping)[1]==0) {
    print("no entrez gene ids mapped")
    return()
  }
  if ( dim(llMapping)[1]>0 & dim(ugMapping)[1]>0 ) {
    merged <- merge(merged, llMapping, by="PROBE", all.x=TRUE)
    merged <- merge(merged, ugMapping, by="PROBE", all.x=TRUE)
    unified <- resolveMaps(merged, trusted=c("EG", "UG"), srcs = c("EG", "UG"))
  }
  if (dim(llMapping)[1]>0 & dim(ugMapping)[1]==0) {
    merged <- merge(merged, llMapping, by="PROBE", all.x=TRUE)
    unified <- resolveMaps(merged, trusted=c("EG"), srcs = c("EG"))
  }
  if (dim(llMapping)[1]==0 & dim(ugMapping)[1]>0) {
    merged <- merge(merged, ugMapping, by="PROBE", all.x=TRUE)
    unified <- resolveMaps(merged, trusted=c("UG"), srcs = c("UG"))
  }

  # unified genemap is now used to collect annotation data form entrez gene (eg)
  parser(eg) <- file.path(.path.package("AnnBuilder"), "scripts", "llParser")
  baseFile(eg) <- unified
  annotation <- parseData(eg, eg@accession, ncol=14)
  colnames(annotation) <- c("PROBE", "ACCNUM", "LOCUSID", "UNIGENE", "GENENAME", "SYMBOL", "CHR", "MAP", "PMID", "GRIF", "SUMFUNC", "GO", "OMIN", "REFSEQ")

  # collect additional chromosome data from golden path

  # load data from golden path with strand <. getStrand(gp)  if gp is the object with th right url
  strand <- getChroLocation(srcUrl(gp), gpLinkNGene())

  # merge new data with current annotation
  annotation <- merge(annotation, strand, by="LOCUSID", all.x=TRUE)

  # collect kegg data
  pways <- findIDNPath(kg) # path LOCUSID <-> PATHNAME
  llPwayMap <- mapLL2ECNPName(kg) # 1. LOCUSID <-> ENZYME; 2. LOCUSID <-> PATH ID

  annotation <- merge(annotation, llPwayMap$llpathname, by.x="LOCUSID", by.y="LOCUSID", all.x=T)
  annotation <- merge(annotation, llPwayMap$llec, by.x="LOCUSID", by.y="LOCUSID", all.x=T)
  annotation <- merge(annotation, pways, by.x="LOCUSID", by.y="LOCUSID", all.x=T)

  # create an R package
  pkgPath <- getwd()
  createEmptyDPkg(packagename, getwd(), force=TRUE)
  annotation <- as.matrix(annotation)
  writeAnnData2Pkg(annotation, pkgName, pkgPath)
  revNames <- intersect(colnames(annotation), c("PMID", "PATH", "ENZYME"))
  if(length(revNames) != 0) {
    writeReverseMap(annotation[ , c("PROBE", revNames)], pkgName, pkgPath)
  }

}


# pathes:
# entrez gene: ftp://ftp.ncbi.nlm.nih.gov/gene/DATA
# uni gene: ftp://ftp.ncbi.nih.gov/repository/UniGene/Hs.data.gz
# golden path: ftp://hgdownload.cse.ucsc.edu/goldenPath/hg17/database   /refLink.txt.gz /refLink.txt.gz
# gene ontology: out of order
# kegg: ftp://ftp.genome.ad.jp/pub/kegg/pathways

