consensusCluster <- function(e, nclass=2, nsamplings=100, npersample=NULL, 
                             hclust.init=TRUE, ...) {
  if (class(e) == "exprSet") e <- exprs(e)
  if (is.null(npersample)) npersample <- floor(0.8*ncol(e))
  e   <- t(e)
  if (hclust.init) {
    hc  <- hclust(dist(e))
    hcc <- cutree(hc,k=nclass)
    init.means <- apply(e,2,function(x)  tapply(x,as.factor(hcc),mean))
    res <- kmeans(e, nclass, centers=init.means, ...)
  } else { res <- kmeans(e, nclass, ...) }
  oo <- order(res$cluster)
  consensCount <- matrix(0, ncol=nrow(e), nrow=nrow(e))
  allCount <- matrix(0, ncol=nrow(e), nrow=nrow(e))
  for (i in 1:nsamplings) {
    if ((i %% 10) == 0) cat(".")
    subsample <- sample(1:nrow(e), replace=FALSE, size=npersample)
    allCount[subsample,subsample] <- allCount[subsample,subsample] + 1
    sub.e <- e[subsample,]
    if (hclust.init) {
      hc  <- hclust(dist(sub.e))
      hcc <- cutree(hc,k=nclass)
      init.means <- apply(sub.e,2,function(x)  tapply(x,as.factor(hcc),mean))
      s <- kmeans(sub.e, nclass, centers=init.means, ...)
    } else { s <- kmeans(sub.e, nclass, ...) }
    for (j in 1:nclass) {
      tmp <- subsample[s$cluster==j]
      consensCount[tmp,tmp] <- consensCount[tmp,tmp] + 1
    }
  }
  cat("\n")
  res$consensusMatrix <- consensCount[oo,oo]/allCount[oo,oo]
  rownames(res$consensusMatrix) <- rownames(e)[oo]
  colnames(res$consensusMatrix) <- rownames(e)[oo]

  res$clusterConsensus <- rep(0, nclass)
  res$itemConsensus <- rep(0, nrow(e))
  names(res$itemConsensus) <- rownames(e)
  for (i in 1:nclass) {
    select <- which(res$cluster[oo]==i)
    m <- res$consensusMatrix[select,select]
    res$clusterConsensus[i] <- (sum(m) - sum(diag(m)))/length(select)/(length(select) - 1)
    for (j in select) {
      res$itemConsensus[j] <- mean(res$consensusMatrix[j,setdiff(select,j)])
    }    
  }

  class(res) <- "consensusCluster"
  return(res)
}

print.consensusCluster <- function(x) {
  cat("Consensus clustering on", sum(x$size), "samples\n")
  cat("  group sizes:", x$size, "\n")
  cat("  consensus: ", x$clusterConsensus, "\n")
}


image.consensusCluster <- function(x, main=NULL, sampleNames=FALSE, outfile=NULL,
                                   pointsize=8, res=72, ...) {
  formNum <- function(x) sapply(x, format, digits=3, nsmall=3)

  nsamples <- length(x$cluster)
  m <- x$consensusMatrix
  leftLabels <- formNum(x$itemConsensus)
  rightLabels <- rownames(m)

  # measure margin widths
  if (is.null(outfile)) X11()
  else bitmap(outfile)
  image(m, axes=FALSE, xlab="cluster consensus", ylab="sample consensus", ...)
  bottomMargin <- 0.9
  leftMargin <- max(strwidth(leftLabels, units="inches")) + 0.7
  if (sampleNames) rightMargin <- max(strwidth(rightLabels)) + 0.5
  else rightMargin <- 0.3
  if (is.null(main)) topMargin <- 0.3 
  else topMargin <- 0.7
  imgHeight <- bottomMargin + nsamples*pointsize/72 +   topMargin
  imgWidth  <-   leftMargin + nsamples*pointsize/72 + rightMargin
  dev.off()

  # draw the image
  if (is.null(outfile)) X11(height=imgHeight, width=imgWidth, pointsize=pointsize)
  else bitmap(outfile, height=imgHeight, width=imgWidth, pointsize=pointsize)
  par(mai=c(bottomMargin, leftMargin, topMargin, rightMargin))
  image(m, axes=FALSE, xlab="cluster consensus", ...)

  # annotations
  if (!is.null(main)) title(main, cex.main=3)

  npercluster <- table(x$cluster)
  thisPos <- 0
  linePos <- NULL
  for (i in 1:(length(npercluster) - 1)) {
    linePos <- c(linePos, npercluster[i]/nsamples)
  }

  axis(side=2, at = ((1:nsamples)-0.5)/nsamples, labels=leftLabels, las=1)
  mtext("sample consensus", side=2, line=4)
  axis(side=1, at = linePos, labels=rep("", length(linePos)))
  if (sampleNames) {
    axis(side=4, at = ((1:nsamples)-0.5)/nsamples, labels=rownames(m),las=1)
  }
  abline(h=linePos,v=linePos)
  thisPos <- 0
  for (i in 1:length(npercluster)) {
    thisPos <- thisPos + npercluster[i]/nsamples/2
    mtext(formNum(x$clusterConsensus[i]),
          side=1, line=1, adj=0.5, at=thisPos, cex=1.5)
    thisPos <- thisPos + npercluster[i]/nsamples/2
  }
  box()
  if (!is.null(outfile)) dev.off()
}











