geneImager <- function(data,
    chip=NULL,
    cluster= "both",     # samples, probes, both, FALSE
    dendro = "both",     # samples, probes, both, FALSE
    center = TRUE,
    scale  = FALSE,      # TRUE/FALSE
    outputPath = dirname(ifelse(is.null(file),"",file)), #
    file   = NULL,       #
    zlim   = NULL,       # c(min,max) L
    annotSamples = NULL, # character vector of length ncol(data)
    annotProbes = NULL,  # character vector of length nrow(data)
    forceExceed,         # T/F (deprecated)
    hide.colorbar = FALSE,
    sample.group = NULL,  # group label for each sample
    group.bgc = c("#666666", "#BBBBBB"), # background color groups
    group.fgc = c("#000000", "#FFFFFF"), # text color groups
    width = NULL,  # if set, width specifies absolute width in inches for whole figure
    height = NULL, # if set, height specifies aboslute height in inches for whole figure, overrides heightScale
    cex.probes = 1, # cex of probes/genes
    cex.samples = 1, # cex of samples
    cex.groups = 1.5, # cex of group labels or vector of cex for each label
    las.groups = 0, # las for each group label (default 0)
    cex.legend = 1.1, # cex of colorbar label and numbers on color bar
DPI = 72, # resolution in dpi
...)
{
# generate a yellow-black-blue color map, slope=1 means linear visual color stretch towards the endpoints, resulting in less black in the middle range
  yb.color <- function(blacklevel=0,slope=2) {
    lsc <- (1:256/256)^(1/slope) 
      lsc=(lsc-min(lsc))/(max(lsc)-min(lsc))*(100-blacklevel)+blacklevel
      yellowH <- 60/360
      blueH <- 240/360
      sat <- lsc/100
      val <- lsc/100
      blacklevel=seq(0,blacklevel,len=256)
      return(c(rgb(r=blacklevel,g=blacklevel,b=rev(lsc),max=100),
	    rgb(r=lsc,g=lsc,b=rev(blacklevel),max=100)))
  }

# convert from inch to centimeter and back
  in2cm <- function(x) {return(x*2.54)}
  cm2in <- function(x) {return(x/2.54)}

# add gene symbols as text string in front of affy ids string vector
  addGeneSymbols <- function(affyIds, affyChip) {
    require(annaffy)
    symbolsToPlot <- aafSymbol(affyIds,affyChip)
      for(i in 1:length(symbolsToPlot)) {
	if(length(unlist(symbolsToPlot[i]))==0) symbolsToPlot[i] = "UNKNOWN"
      }
    return(paste(symbolsToPlot," (",affyIds,")",sep=""))
  }

# rewrote of strheight to support descenders
  strheight <- function(s,...) {
    origStrheight=get("strheight","package:graphics");
    return(origStrheight(s,...)*ifelse(length(grep("[jgy_pq]",s)),1.28,1))
  }

# calculate sample group height in inches
  groupHeight <- function(sample.group,cex.groups,las.groups) {
    gHeight=0;
    if (!is.null(sample.group)) {
      cls <- table(sample.group)
	cls <- cls[as.character(unique(sample.group))] 
	for(i in 1:length(cls)) {
	  if (length(las.groups)==1) l=las.groups else l=las.groups[i]
	    if (length(cex.groups)==1) c=cex.groups else c=cex.groups[i]
	      if (l>1)
		gHeight=max(gHeight,strwidth(names(cls[i]),units="inch",cex=c/par("cex"))*yinch(1)/xinch(1)*2)
	      else
		gHeight=max(gHeight,strheight(names(cls[i]),units="inch",cex=c/par("cex")))
	}
    }
    return(gHeight)
  }


  # errors
  if (!file.exists(outputPath)) stop("Output path does not exist", outputPath)
  if (cluster!=FALSE && !pmatch(cluster,c("samples","probes","both"),nomatch=FALSE)) { 
	stop("Invalid cluster option! (none of [FALSE | 'samples' | 'probes' | 'both']") 
      }
  if (!is.null(file)) {
    if ((dirname(file) != outputPath) & (dirname(file) != "."))
      stop("'outputPath' and directory of 'file' unequal")
    fullFile <- file.path(outputPath, basename(file))
  }

  # warnings
  if (!missing(forceExceed)) warning("'forceExceed' is deprecated")
#  if (cluster!=FALSE && pmatch(cluster,c("samples","both"),nomatch=FALSE)) {
#    if (!is.null(sample.group)) {
#      sample.group <- NULL
#      warning("sample.group overwritten by cluster=", cluster)
#    }
#  }
  if (dendro!=FALSE) {
    if (!pmatch(dendro,c("samples","probes","both"),nomatch=FALSE)) { 
    stop("Invalid dendro option! (none of [FALSE | 'samples' | 'probes' | 'both']")
    }
    if (pmatch(dendro,c("samples","both"),nomatch=FALSE) && pmatch(cluster,"probes",nomatch=FALSE)) {
     stop("Invalid dendro option when cluster=probes dendro can only be probes ")
    }
    if (pmatch(dendro,c("probes","both"),nomatch=FALSE) && pmatch(cluster,"samples",nomatch=FALSE)) {
     stop("Invalid dendro option when cluster=samples dendro can only be samples")
    }
  }

  anchors <- annotProbes
  if (is.logical(annotProbes)) {
    if (annotProbes) {
      anchors   <- rownames(data)
	if (is.null(chip)) {
	  annotProbes <- anchors
	} else {
	  annotProbes <- addGeneSymbols(anchors,chip)
	}
      if (is.null(annotProbes) || length(annotProbes) != nrow(data)) 
	stop("No probe annotation provided.")
    } else annotProbes <- NULL
  }
  if (!is.null(names(annotProbes))) {
    anchors <- names(annotProbes)
  } 

  if (is.logical(annotSamples)) {
    if (annotSamples) {
      annotSamples <- colnames(data)
    } else annotSamples <- NULL
  }
  if (!is.null(sample.group)) {
    if (!is.null(annotSamples) && length(annotSamples) != length(sample.group)) 
      warning("One group label per sample required")
    glevels=as.numeric(as.factor(sample.group))
# test if all groups are already grouped together if not, sort them
    if(sum(duplicated(glevels[c(T,glevels[-1]!=glevels[-length(glevels)])]))) {
      sampleOrder <- order(sample.group)
      data <- data[,sampleOrder, drop=FALSE]
      sample.group <- sample.group[sampleOrder]
      if (!is.null(annotSamples)) { 
        annotSamples <- annotSamples[sampleOrder]
      }
    }
  }

# scale data
  if (center) {
    if (scale) { data <- t(scale(t(data)))
    } else { data <- t(scale(t(data),scale=FALSE)) }
  } else {
    if (scale) warning("scaling is ignored since centering is turned off.")
  }

# cluster
  if (ncol(data) > 1 & (cluster == "samples" | cluster == "both")) {
    if (is.null(sample.group)) { cluster.group <- rep("one", ncol(data))
    } else { cluster.group <- sample.group }
    for (this.group in unique(cluster.group)) {
      select <- cluster.group == this.group
      sample_dist <- dist(t(data[, select, drop=FALSE]))
      sample_clust <- hclust(sample_dist)
      data[,select] <- data[, select, drop=FALSE][, sample_clust$order, drop=FALSE]
      if (!is.null(annotSamples)) { 
	annotSamples[select] <- annotSamples[select][sample_clust$order]
      }
    }
  }

  if (nrow(data) > 1 && (cluster == "probes" || cluster == "both")) {
    gene_dist <- dist(data)
      gene_clust <- hclust(gene_dist)
      data   <- data[gene_clust$order, , drop=FALSE]
      annotProbes <- annotProbes[gene_clust$order]
      anchors     <- anchors[gene_clust$order]
  }

# define maximal widthHeatmap/heightHeatmap area
  if (is.null(file)) { 
    x11()
par("ps"=16)
  } else {
    bitmap(file=fullFile,type="png16m",...)
par("ps"=16)
  }
#cat("empty plot\n")

  widthColorbar=widthLeftDendro=heightTopDendro=0;
  layout(matrix(c(0,0,0,0,0,1),2,3,byrow=T));
  plot(0);

  if (!is.null(annotSamples)) {
# device play ground
    heightSampleLabel<-in2cm(max(strwidth(annotSamples,units="inches",cex=cex.samples/par("cex"))))
  } else {
    heightSampleLabel<-1
  }

  if (!is.null(annotSamples)) {
    widthHeatmap       <- in2cm(sum(strheight(annotSamples,units="inches",cex=cex.samples/par("cex")))) #cl,cex=cex.samples
  } else {
    widthHeatmap       <- max(3,ncol(data)/4)
  }
  if (!is.null(width)) widthHeatmap=in2cm(width/DPI);
  MAX_PROBES_DISPLAY=100 # maximal number of probes where we would display each probe in full text height (HTML anchors are correct then)
  DEFAULT_HEIGHT=10  # default height of heatmap if more probes than MAX_PROBES_DISPLAY
  if (is.null(height) && !is.null(annotProbes)) 
    heightHeatmap<- in2cm(max(1,nrow(data)/30,sum(strheight(annotProbes,units="inches",cex=cex.probes/par("cex"))))) 
  else heightHeatmap <- in2cm(ifelse(is.null(height),ifelse(nrow(data)>MAX_PROBES_DISPLAY,DEFAULT_HEIGHT,strheight("A",units="inches",cex=cex.probes/par("cex"))*nrow(data)),height/DPI));

  if(!is.null(annotProbes)) {
    if (nrow(data) == 1) {
      ats <- 0.5
      origAts <- ats
      nonEmpty=1
    } else {
      nonEmpty=which(annotProbes!="") # only draw tickmarks when labels present
      ats=(nonEmpty-1)/(nrow(data)-1) # where should the labels be drawn?
# better try to arrange labels so that they do not overlap
      origAts=ats
      percentLabeled=length(nonEmpty)/length(annotProbes)
      labelHeight=yinch(max(strheight(annotProbes,units="inch",cex=cex.probes/par("cex"))))
#      rowHeightIn=cm2in(heightHeatmap)*ats[2]
#      rowHeightIn=yinch(ats[2])
#      if(percentLabeled<0.7) rowHeightIn=labelHeight/1.3
#      else rowHeightIn=0.008*(1-percentLabeled)/0.3/cex.probes;
#labelConflict=(ats[-1]-ats[-20])<rowHeightIn # where are the conflicts?
      if (percentLabeled<0.95) { # is there enough space to arrange labels better?
        overlapScale=15
        while(TRUE) {
          labelConflicts=vector()
          conflicts=0;
          labelConflicts[length(ats)]=0
          for(i in length(ats):2) {
            if ((ats[i]-ats[i-1])<labelHeight/overlapScale) conflicts=conflicts+1 else conflicts=0;
            labelConflicts[i-1]=conflicts;
          }
          middle=0;
          for(i in 1:length(ats)) {
            if (labelConflicts[i]>0 || middle>0) {
              if (middle==0) middle=labelConflicts[i]/2;
              shiftLen=labelHeight/(3*overlapScale);
              if (labelConflicts[i]>middle) ats[i]=ats[i]-shiftLen else 
                if (labelConflicts[i]<middle) ats[i]=ats[i]+shiftLen;
              if (labelConflicts[i]==0) middle=0;
            } 
          }
          if (sum(labelConflicts)==0) break;
        } 
      }
    }
  }

      heightGroupSample=groupHeight(sample.group,cex.groups,las.groups)
      groupSampleBoxStretch=1.4
  heightSampleGroupLabel <- in2cm(ifelse(is.null(sample.group),0,heightGroupSample))*groupSampleBoxStretch
# width probe annotation 
# do we use axis tickmarks or paint our own zickzack tickmarks
    widthProbeLabel<- ifelse(is.null(annotProbes),0.5,in2cm(max(strwidth(annotProbes,units="inches",cex=cex.probes))+ifelse(identical(ats,origAts),1,1.2)*strwidth("A",units="inches",cex=cex.probes/par("cex"))))

#  cat("height_top: ", heightTopDendro," height_var: ",heightHeatmap, " height_bottom: ", heightSampleLabel,"\n",sep="")
    if (is.null(zlim)) {
      zMin <- -(max(abs(min(data)),abs(max(data))))
	zMax <- -zMin

	if (scale) {
	  zlim <- c(-3,3)
	    data[data > 3] <- 3
	    data[data < (-3)] <- -3
	} else {
	  minV <- quantile(data,0.025)
	    maxV <- quantile(data,0.975)
	    minVg <- -max(abs(minV),abs(maxV))
	    maxVg <- max(abs(minV),abs(maxV))
	    zlim <- c(minVg,maxVg)
	}
    }

  lv <- c(0,0,0,0,1,0) # [2]:upper dendro, [4]:left dendro, [5]:1=heatmap, [6]:colorbar
    nrPlots=1
    if (!hide.colorbar) {
      nrPlots=nrPlots+1
	lv[6]=nrPlots;
      if (scale) {
	tickLabels=signif(seq(zlim[1],zlim[2],length=5),2)
	  colorBarLabel="Sigma (standardized)" 
      } else  {
        if (center) { 
          tickLabels=round(exp(seq(zlim[1],zlim[2],length=5)),2)
          colorBarLabel="Fold Change" #with respect to genewise mean over all patients
        } else { 
          tickLabels=round(seq(zlim[1],zlim[2],length=5),2)
          colorBarLabel="" 
        }
      }
      widthLegendLabel=in2cm(max(strheight(tickLabels,units="inches",cex=cex.legend/par("cex")))+strheight(colorBarLabel,units="inches",cex=cex.legend/par("cex")))
	widthColorbar=widthLegendLabel+0.5+0.3;
    }

  if (dendro == "both" || dendro == "samples" )  {
    if (cluster == "both" || cluster == "samples") {
      heightTopDendro=3;
      nrPlots=nrPlots+1
	lv[2]=nrPlots;
    } else {
      cat("WARNING: Not able to plot dendrograms. Please set: 'cluster=\"both\" or \"samples\"'.\n")
	dendro <- FALSE
    }
  }
  if (dendro == "both" || dendro == "probes" )  {
    if (cluster == "both" || cluster == "samples") {
      widthLeftDendro=3;
      nrPlots=nrPlots+1
	lv[4]=nrPlots;
    } else {
      cat("WARNING: Not able to plot dendrograms. Please set: 'cluster=\"both\" or \"probes\"'.\n")
	dendro <- FALSE
    }
  }
  zoom=1
    if (!is.null(file)) {
      zoom=1/1
      zoom2=1 #0.85
	heightHeatmap=heightHeatmap/zoom;
      widthHeatmap=widthHeatmap/zoom;
      widthProbeLabel=widthProbeLabel/zoom2
      heightSampleLabel=heightSampleLabel/zoom2
    }
  lwidth <- c(widthLeftDendro,widthHeatmap+widthProbeLabel,widthColorbar)
    lheight <- c(heightTopDendro,heightSampleLabel+heightHeatmap+heightSampleGroupLabel)

    dev.off()
    if (is.null(file)) {
      x11(width=cm2in(sum(lwidth)+.5),height=cm2in(sum(lheight)+.5))
par("ps"=16)
    } else {
      bitmap(file=fullFile, type="png16m",width=cm2in(sum(lwidth)+1),height=cm2in(sum(lheight)+1),...)
par("ps"=16)
    }
  pl <- layout(matrix(lv,2,3,byrow=T),lcm(lwidth),lcm(lheight))

    data[data > zlim[2]] <- zlim[2]
    data[data < zlim[1]] <- zlim[1]

#par(mar=c(0,0,0,0))
#  par(omi=c(0,0,0,0))

# heatmap
    par(mai=cm2in(c(heightSampleLabel,0,heightSampleGroupLabel,widthProbeLabel)))

#cat("heat map\n")
    image(t(data),xaxt='n',yaxt='n',,col=yb.color(TRUE),zlim=zlim)
    box()

    if (!is.null(annotSamples) & ncol(data) > 1) {
# cat("sample annotation\n")
#      par("ps"=15)
	axis(1,at=(0:(ncol(data)-1))/(ncol(data)-1),annotSamples,las=3,cex.axis=cex.samples);
    }

#cat("probe annotation\n")
#axis(4,at=(0:(nrow(data)-1))/(nrow(data)-1),annotProbes,las=1,cex.axis=cex.probes);
#  par("ps"=15)


    if(!is.null(annotProbes) ){
      if (identical(ats,origAts)) {
	axis(4,at=ats,annotProbes[nonEmpty],las=1,cex.axis=cex.probes)
      } else {
#mtext(side=4,line=2.5,at=ats,annotProbes[nonEmpty],las=1,cex=par("cex")*par("cex.axis"))
	tickmarkLen=xinch(strwidth("A",units="inches",cex=cex.probes/par("cex")))/1.1
	  tickmarkStart=1+0.5/(ncol(data)-1)
	  mtext(side=4,line=2,at=ats,annotProbes[nonEmpty],las=1,cex=cex.probes*par("cex"))
	  for(i in 1:length(ats)) {
	    x1 <- tickmarkStart; 
	    x2 <- tickmarkStart+tickmarkLen*0.2;
	    x3 <- tickmarkStart+tickmarkLen*0.8;
	    x4 <- tickmarkStart+tickmarkLen;
	    y1 <- y2 <- origAts[i]
	      y3 <- y4 <- ats[i]
	      lines(c(x1,x2,x3,x4),c(y1,y2,y3,y4),xpd=NA)
	  }
      }
    }
  if (!is.null(sample.group)) {
    cls <- table(sample.group)
      cls <- cls[as.character(unique(sample.group))] # cls[order(names(cls))]
      ybot <- 1 + 0.5/(nrow(data)-1)
      ytop <- ybot + yinch(heightGroupSample)*groupSampleBoxStretch
      tmp <- -0.5; lblAt <- vector()
      for (i in 1:length(cls)) {
	lblAt[i] <- tmp + cls[i]/2
	  rect(tmp/(ncol(data)-1), ybot, (tmp+cls[i])/(ncol(data)-1), ytop, 
	      col=group.bgc[i%%length(group.bgc)+1], border=TRUE, xpd=NA)
	  tmp <- tmp + cls[i]
      }
#axis(3, lblAt/(ncol(data)-1), names(cls),cex.axis=2)
    for(i in 1:length(lblAt)) {
      if (length(las.groups)!=1) l=las.groups[i]  else l=las.groups;
      if (!is.null(cex.groups)) if (length(cex.groups)==1) c=cex.groups else c=cex.groups[i] else c=1;
      mtextLineHeight=yinch(par("mai")[3]/par("mar")[3])  # height of one mtext line in inches
	mtextLineOffset=yinch(par("csi")/par("cra")[2]*2) # this is the offset in inches for line=0 in mtext, exactly two pixels

 descendersAdd=ifelse(length(grep("[jgy_pq]",names(cls))),strheight(names(cls)[i])/3/mtextLineHeight,0) # shift group labels a bit up if we have descenders
	mtext(side=3, line=descendersAdd+(yinch(heightGroupSample)*(groupSampleBoxStretch-1)/2-mtextLineOffset)/mtextLineHeight, at=(lblAt/(ncol(data)-1))[i], text=names(cls)[i], cex=c, col=group.fgc[(i-1)%%length(group.fgc)+1],las=l) # put text exactly in the middle
    }
#    mtext(side=3, line=0.6, at=lblAt/(ncol(data)-1), text=names(cls), cex=1.3, col=group.fgc)
    cls <- cls[-length(cls)] # toss the last group
      if (length(cls)>1) for(i in 2:length(cls)) cls[i]=cls[i]+cls[i-1]
	abline(v=(cls-0.5)/(ncol(data)-1), col="white", lty=3)
  }

#cat("empty plot\n")
#  if (!is.null(annotSamples)) {
#    par("mai"=c(0,0,0,0))
#    plot.new()
#  }

#  source("~/compdiagPipeline/r_packages/compdiagTools/R/geneImager.R"); geneImager(matrix(rnorm(100),10,10,dimnames=list(paste("gene",1:10),paste("sample",1:10))), annotProbes=T, scale=TRUE,annotSamples=T)

    if (!hide.colorbar) {
# legend
#cat("plot legend\n")
      if (is.null(file)) { par(mai=cm2in(c(heightSampleLabel,0,heightSampleGroupLabel,widthLegendLabel)))
      } else { par(mai=cm2in(c(heightSampleLabel,0,heightSampleGroupLabel,widthLegendLabel))) }
      image(matrix(seq(zlim[1],zlim[2],length=100),1,100),xaxt='n',yaxt='n', col=yb.color(TRUE),zlim=zlim)
	box()

zoom=0.5
	axis(4,at=0:4/4,tickLabels,las=3,cex.axis=cex.legend,mgp=c(3,0.8*cex.legend/zoom,0))
	mtext(colorBarLabel,side=4,line=2*cex.legend/zoom,cex=cex.legend*par("cex"))
    }
# dendrogram sample
  if (dendro == "samples" | dendro == "both") {
    if (cluster == "samples" | cluster == "both") {
      if (ncol(data) > 2) {
	par(mai=cm2in(c(heightSampleGroupLabel,0,0,widthProbeLabel)))
	  plot(as.dendrogram(sample_clust),horiz=FALSE,leaflab="none",axes=FALSE,xaxs = "i")
      }
    }
  }

# dendrogram probes
#cat("plot gene cluster\n")
  if (dendro == "probes" | dendro == "both") {
    if (cluster == "probes" | cluster == "both") {
      if (nrow(data) > 2) {
	par(mai=cm2in(c(heightSampleLabel,0,heightSampleGroupLabel,0)))
	  plot(as.dendrogram(gene_clust),horiz=TRUE,leaflab="none",axes=FALSE,yaxs = "i")
#if (dendro=="probes") { plot.new(); par("mfg"=c(1,2)); plot(0,type="n",axes=F,xlab="",ylab="")} # hack as otherwise the next image crashes for strange reasons
#if (dendro=="both") { oldmfg=par("mfg"); plot.new(); par("mfg"=oldmfg) } # this is quite a hack but otherwise the next image crashes for strange reasons
      }
    }
  }



  if (!is.null(file)) { dev.off() }

  if (!is.null(file)) {
    if (!is.null(annotProbes)) {
      if (dendro == "probes" | dendro == "both") { x_offset <- widthLeftDendro + widthHeatmap + 0.25
      } else { x_offset <- widthHeatmap + 0.25 }
      if (dendro == "samples" | dendro == "both") { y_offset <- heightTopDendro + 0.1
      } else { y_offset <- 0.1 }
      if (!is.null(sample.group)) y_offset <- y_offset + heightSampleGroupLabel

	  map_width  <- widthProbeLabel
	  map_height <- heightHeatmap
	  x_offset.px   <- round(cm2in(x_offset)  * DPI)
	  y_offset.px   <- round(cm2in(y_offset)  * DPI)
	  map_width.px  <- round(cm2in(map_width) * DPI)
	  map_height.px <- round(cm2in(map_height)* DPI)
	  mapstr <- "<map name=probemap>\n"

	  y_step     <- map_height.px / nrow(data)
	  y_step.r   <- round(y_step)
	  step_error <- y_step - y_step.r

	  cum_step_error = 0
	  y1 <- y_offset.px+5
	  for (i in 1:nrow(data)) {
	    x1 <- x_offset.px
	      x2 <- x_offset.px + map_width.px

	      cum_step_error = cum_step_error + step_error;
	    if (cum_step_error > 1) {
	      y1 = y1+1
		cum_step_error = 0;
	    }

	    if (cum_step_error < -1){
	      y1 = y1-1
		cum_step_error=0;
	    }

	    y2 <- y1 + y_step.r
	      mapstr <- paste(mapstr,"<area shape=\"rect\" coords=\"", x1,
		  ",",y1,",",x2,",",y2,"\" href=\"#",anchors[nrow(data)-i+1],
		  "\" alt=\"",annotProbes[nrow(data)-i+1],"\">\n",sep="")
	      y1 <- y2
	  }

      mapstr <- paste(mapstr,"</map>\n<img usemap=\"#probemap\" src=\"",file,
	  "\" border=0>\n",sep="")
    } else {
      mapstr <- paste("<img src=\"",file,"\" border=0>\n",sep="")
    }
    return(invisible(mapstr))
  } else {
    return(invisible(NULL))
  }
}
