aafGetHTML <- function(tab, pos, textProperty, urlProperty=NULL, linkText=NULL) {
  # generates HTML for 'textProperty' at row 'pos' in
  # aafTable 'tab'. 'linkText' is used instead of 
  # text extracted from data packages, if it is not
  # NULL. If urlProperty is provided, the url is extracted
  # from that property in the same row. In this latter 
  # case, text and url are substituted if some element is not
  # available.
  x <- tab[pos, which(colnames(tab)==textProperty)]@table[[1]]
  if (!is.null(urlProperty)) u <- tab[pos, which(colnames(tab)==urlProperty)]@table[[1]]
  else u <- NULL

  # find data
  if (length(x) == 0) {
    if (length(u) == 0) { return("n/a")
    } else { 
      itemText <- sapply(u, getText) 
      itemUrl <- sapply(u, getURL)
    }
  } else { 
    itemText <- sapply(x, getText) 
    if (length(u) == 0) { itemUrl <- sapply(x, getURL)
    } else { itemUrl <- sapply(u, getURL) }
  }
  if (!is.null(linkText)) text <- linkText

  # generate HTML
  if (length(itemUrl) == 0) return(paste(itemText, collapse=", "))
  if (is.null(linkText)) {
    return(paste("<a href='", itemUrl, "' target='annotation'>", itemText, "</a>", 
                 sep="", collapse=", "))
  }
  tmp <- paste("<a href='", itemUrl, "' target='annotation'>", linkText, "</a>", 
               sep="", collapse=", ")
  return(paste("[", tmp, "]"))
}

logDataPackageVersion <- function(pkgs) {
  if (is.null(pkgs)) tmp <- "<li>No annotation data available</li>"
  else {
    tmp <- paste("<li>Data package '", pkgs, "' version ", 
                 sapply(pkgs, function(pkg) packageDescription(pkg)[["Version"]]), 
                 " packaged on ", 
                 sapply(pkgs, function(pkg) packageDescription(pkg)[["Packaged"]]), 
                 "</li>", sep="", collapse="")
  }
  return(paste("<p>Annotations from:<ul>", tmp, "</ul></p>", sep=""))
}

geneLister <- function(obj,
		       chip=NULL,
		       hc="#99CCCC",
		       bgc1="#CCCCCC",
		       bgc2="#E8E8E8",
		       HTML           = TRUE,
		       info	      = NULL,
		       annotCORG      = FALSE,
		       annotLOCUS     = TRUE,
		       annotGO        = TRUE,
		       annotSYMBOL    = TRUE,
		       annotGDESCRIPTION = TRUE,
		       annotPROBE        = TRUE,
		       annotUNIGENE   = TRUE,
		       annotGENBANK   = TRUE,
		       annotLOCUSLINK = TRUE,
		       annotPUBMED    = TRUE,
		       annotKEGG      = TRUE,
		       annotOMIM      = TRUE,
		       annotGENECARDS = TRUE,
		       sep	      = "\t",
		       verbose	      = FALSE,
		       maxfdr	      = 1,
		       maxgenes	      = 500,
                       lookupProbes   = NULL,
		       groupname      = "n/a",
		       links          = FALSE) {
  require(annaffy) || stop("Package annaffy unavailable")

  ##############################################
  # define external links

  OMIMlink.raw  <- "http://www.ncbi.nlm.nih.gov:80/entrez/dispomim.cgi?id="
  GClink.raw    <- "http://bioinfo.weizmann.ac.il/cards-bin/cardsearch.pl?search=_s_"

  ##############################################
  s <- NULL
  v <- attributes(obj)$class[1]=="geneList"
  addInformation = NULL

  fdrTable <- NULL
  if (length(v)>0) {
    if (attributes(obj)$class[1] == "twilight") {
      # compute n-genes for some fdrs
      fdrTable <- c(sum(obj$result[,"qvalue"] <= 0.001),
                        sum(obj$result[,"qvalue"] <= 0.01),
                        sum(obj$result[,"qvalue"] <= 0.05),
                        sum(obj$result[,"qvalue"] <= 0.1),
                        sum(obj$result[,"qvalue"] <= 0.2))

      # order twilight object
      ord <- order(obj$result[,"qvalue"],-abs(obj$result[,"observed"]))
      obj$result <- obj$result[ord,]

      obj$result <- obj$result[obj$result[,"qvalue"]<=maxfdr,]
      if (nrow(obj$result)>maxgenes) {
        obj$result <- obj$result[1:maxgenes,]
      }

      if (nrow(obj$result)==0) {
        stop("Empty geneList. No probes survived FDR filtering.\n");
      }

      fdr <- obj$result[nrow(obj$result),"qvalue"]
  
      tmp <- obj$result[obj$result[,"observed"] > 0,]
      upregulated   <- data.frame("Probe"=rownames(tmp), row.names=rownames(tmp))
      upregulated[["observed"]]  <- round(tmp[,"observed"], 3)
      upregulated[["q.value"]]   <- format.pval(tmp[,"qvalue"], digits=3)
      upregulated[["local.fdr"]] <- format.pval(tmp[,"fdr"], digits=3)

      tmp <- obj$result[obj$result[,"observed"] < 0,]
      downregulated <- data.frame("Probes"=rownames(tmp), row.names=rownames(tmp))
      downregulated[["observed"]] <- round(tmp[,"observed"], 3)
      downregulated[["q.value"]]  <- format.pval(tmp[,"qvalue"], digits=3)
      downregulated[["local.fdr"]] <- format.pval(tmp[,"fdr"], digits=3)

      nup   <- nrow(upregulated);
      ndown <- nrow(downregulated);

      if (is.null(info)) {
        info=c("observed","q.value", "local.fdr")
      }

      gL_up <- new("geneList")
      gL_up@description <- "dummy List"
      gL_up@creator     <- "twilight"
      gL_up@parameter   <- ""
      gL_up@groups      <- ""
      gL_up@link        <- ""
      gL_up@chip        <- ifelse(is.null(chip), "", chip)
      gL_up@geneInfos   <- upregulated

      gL_dn <- new("geneList")
      gL_dn@description <- "dummy List"
      gL_dn@creator     <- "twilight"
      gL_dn@parameter   <- ""
      gL_dn@groups      <- ""
      gL_dn@link        <- ""
      gL_dn@chip        <- ifelse(is.null(chip), "", chip)
      gL_dn@geneInfos   <- downregulated

      s <- paste("<ul>\n<li> False discovery rate: ",format.pval(fdr),"</li>\n<li>",nup,
                 " <a href=\"#upreg\">Upregulated</a> genes in group ", groupname,"</li>\n<li>", 
                 ndown, " <a href=\"#downreg\">Downregulated</a> genes in group ", 
                 groupname, "</li>\n</ul>\n", sep="")

      if (!is.null(fdrTable)) {
        s <- paste(s, "<P><UL><UL><TABLE CELLSPACING='5'>",
                   "<TR ALIGN=CENTER BGCOLOR='#99CCCC'><TH>FDR</TH>",
                   "<TH>0.001</TH><TH>0.01</TH><TH>0.05</TH><TH>0.1</TH>",
                   "<TH>0.2</TH></TR><TR ALIGN=CENTER BGCOLOR='#E8E8E8'>",
                   "<TH BGCOLOR='99CCCC'>#Genes</TH>",
                   paste("<TD>", fdrTable, "</TD>", sep="", collapse=""),
                   "</TR></TABLE></UL></UL></P>", sep="")
      }
      s<-paste(s,"<a name='upreg'></a><h3>Upregulated genes of group ",groupname,"</h3>",
               "\n<ul>\n<li>to <a href=\"#downreg\">downregulated</a> genes of ", groupname,
               "</li>\n</ul>\n",
               geneLister(gL_up, chip, hc, bgc1, bgc2, HTML, info, 
                          annotCORG, annotLOCUS, annotGO, annotSYMBOL, annotGDESCRIPTION, 
                          annotPROBE, annotUNIGENE, annotGENBANK, annotLOCUSLINK, 
                          annotPUBMED, annotKEGG, annotOMIM, annotGENECARDS, sep, verbose,
                          maxfdr, maxgenes, lookupProbes, groupname, links),
               "<br><br><a name=\"downreg\"></a><h3>Downregulated genes of group ", 
               groupname, "</h3>\n<ul>\n",
               "<li>to <a href=\"#upreg\">upregulated</a> genes of ", groupname, 
               "</li>\n</ul>\n",
               geneLister(gL_dn, chip, hc, bgc1, bgc2, HTML, info, 
                          annotCORG, annotLOCUS, annotGO, annotSYMBOL, annotGDESCRIPTION,
                          annotPROBE, annotUNIGENE, annotGENBANK, annotLOCUSLINK, 
                          annotPUBMED, annotKEGG, annotOMIM, annotGENECARDS, sep,verbose,
                          maxfdr, maxgenes, lookupProbes, groupname, links), 
               logDataPackageVersion(chip),
               sep="")
      return(s)
    }
  }

  if (length(v)>0) {
    # it is a gene list ...
    probes <- rownames(obj@geneInfos);
    chip   <- obj@chip;

    # obtain additional list information
    if (length(info)) {
      infColumns <- colnames(obj@geneInfos) %in% as.vector(info)

      if (sum(infColumns)>0) {
        addInformation <- obj@geneInfos[,infColumns,drop=FALSE]
      }
    }
  } else {
    # it's something else
    if (is.null(chip)) {
      stop("geneLister: No Chip defined!")
    }
    probes <- obj;
  }

  if (!require(chip,character.only = TRUE)) {
    warning("no annotation data found for supplied chip\n");
    return("")
  } else {

    ##################################################
    # check existence of annotation packages

    packages <- ls(paste("package:",chip,sep=""))

    if (!(paste(chip,"MAP",sep="") %in% packages)) {annotLOCUS=FALSE}
    if (!(paste(chip,"GO",sep="") %in% packages)) {annotGO=FALSE}
    if (!(paste(chip,"SYMBOL",sep="") %in% packages)) {annotSYMBOL=FALSE}
    if (!(paste(chip,"GENENAME",sep="") %in% packages)) {annotGDESCRIPTION=FALSE}
    if (!(paste(chip,"UNIGENE",sep="") %in% packages)) {annotUNIGENE=FALSE}
    if (!(paste(chip,"ACCNUM",sep="") %in% packages)) {annotGENBANK=FALSE}
    if (!(paste(chip,"LOCUSID",sep="") %in% packages)) {annotLOCUSLINK=FALSE}
    if (!(paste(chip,"PMID",sep="") %in% packages)) {annotPUBMED=FALSE}
    if (!(paste(chip,"PATH",sep="") %in% packages)) {annotKEGG=FALSE}
    if (!(paste(chip,"OMIM",sep="") %in% packages)) {annotOMIM=FALSE}

    if (annotCORG) {
      if (!require(CorgAnn)) {
        annotCORG=FALSE
        cat("geneLister: WARNING! CORG annotation package not found\n")
      } else {
        options(warn=-1)
        crgPackages <- ls(package:CorgAnn)
        options(warn=0)
        if (!(paste(chip,"CORG",sep="") %in% crgPackages)){
          annotCORG=FALSE
          cat("geneLister: WARNING! no CORG annotation for desired chip available\n")
        }
      }
    }
  }

  if (length(probes)>0) {
    if (is.null(lookupProbes)) lookupProbes <- probes
    aafColumns <- NULL
    if (annotSYMBOL)       { aafColumns <- c(aafColumns, "Symbol") }
    if (annotGDESCRIPTION) { aafColumns <- c(aafColumns, "Description") }
    if (annotLOCUS) { aafColumns <- c(aafColumns, "Cytoband") }
    if (annotUNIGENE)      { aafColumns <- c(aafColumns, "UniGene") }
    if (annotGENBANK)      { aafColumns <- c(aafColumns, "GenBank") }
    if (annotLOCUSLINK)    { aafColumns <- c(aafColumns, "LocusLink") }
    if (annotPUBMED)       { aafColumns <- c(aafColumns, "PubMed") }
    if (annotGO)       { aafColumns <- c(aafColumns, "Gene Ontology") }
    if (annotKEGG)       { aafColumns <- c(aafColumns, "Pathway") }
    aafTab <- aafTableAnn(lookupProbes, chip, aafColumns)
  }

  # HTML part or non HTML part
  if (HTML) {
    n <- length(probes)
    if (n>0) {
      headers <- c()
      if (!is.null(addInformation)) headers <- c(headers, "Info")
      if (annotPROBE | annotSYMBOL | annotLOCUS | annotUNIGENE)
        headers <- c(headers, "Probe Set")
      if (annotGDESCRIPTION)        headers <- c(headers, "Description")
      if (annotGO)                  headers <- c(headers, "Gene Ontology")
      if (annotCORG)                headers <- c(headers, "CORG")
      if (annotGENBANK | annotPUBMED | annotKEGG | annotOMIM | annotGENECARDS) 
        headers <- c(headers, "Misc")

      htmlTable <- matrix("", nrow=n, ncol=length(headers))

      for (i in 1:n) {
        j <- 0 # column counter

        # additional information block
        if (!is.null(addInformation)) {
          j <- j + 1
          tmp <- paste("<tr><td><b>", colnames(addInformation), ": </b></td>\n<td>",
                       addInformation[probes[i],], "</td>\n</tr>\n", sep="", collapse="")
          htmlTable[i, j] <- paste("\n<table border=0>\n", tmp, "</table>\n", sep="")
        }

        # probe set information
        if (annotPROBE | annotSYMBOL | annotLOCUS | annotUNIGENE) {
          j <- j + 1
          tmpHTML <- NULL
          
          if (annotPROBE) {
            # probe ids shown on a indexpage should link on the
            # probe anchors on the main result page. Such links are 
            # generated when the parameter links is TRUE.
            if(links) {
              probeUrl <- paste("../index.html\#", probes[i], sep="")
              target <-  "index"
            } else {
              probeUrl <- getURL(aafProbe(lookupProbes[i]))
              target <- "annotation"
            }
            probeText <- gsub("/", " / ", probes[i]) # allow word wrap
            tmpHTML <- c(tmpHTML, paste("<a href='", probeUrl, "' target='", target, "'>",
                                        probeText, "</a>", sep=""))
           }

          # symbol information with entrez gene link
          if (annotSYMBOL) {
            if (annotLOCUSLINK) {
              symbolHTML <- aafGetHTML(aafTab, i, "Symbol", "LocusLink")
            } else {
              symbolHTML <- aafGetHTML(aafTab, i, "Symbol")
            }
            tmpHTML <- c(tmpHTML, symbolHTML)
          }

          # UniGene link
          if (annotUNIGENE) {
            unigeneHTML <- aafGetHTML(aafTab, i, "UniGene")
            tmpHTML <- c(tmpHTML, unigeneHTML)
          }

          # annot ytoband 
          if (annotLOCUS) {
            cytoHTML <- aafGetHTML(aafTab, i, "Cytoband")
            tmpHTML <- c(tmpHTML, cytoHTML)
          }

          htmlTable[i, j] <- paste(tmpHTML, collapse="<BR>")
        }

        # annot GDESCRIPTION information
        if (annotGDESCRIPTION) {
          j <- j + 1
          htmlTable[i, j] <- aafGetHTML(aafTab, i, "Description")
        }

        # annot GO information
        if (annotGO) {
          j <- j + 1
          htmlTable[i, j] <- getGOTable(aafTab, i)
        }

        # annot CORG information
        if (annotCORG) {
          j <- j + 1
          p <- as.name(paste(chip, "CORG", sep=""))
          if (exists(lookupProbes[i],eval(p))) { corgTF <- get(lookupProbes[i],eval(p)) }
          else { corgTF <- NA }
          if(!is.na(corgTF[1])) { 
            htmlTable[i, j] <- paste(corgTF, collapse=", ")
          } else {
            htmlTable[i, j] <- "NA"
          }
        }

        if (annotGENBANK | annotPUBMED | annotKEGG | annotOMIM | annotGENECARDS) {
          j <- j + 1
          tmp <- "\n<table border=0>\n"

          # REFER TO getMiscTable.R
          if(annotGENBANK) {
            tmp <- paste(tmp, "<tr valign=top>\n<td>GenBank:</td>\n<td>", 
                         aafGetHTML(aafTab, i, "GenBank"), "</td>\n</tr>\n", sep="")
          }
          if(annotPUBMED) {
            tmp <- paste(tmp, "<tr valign=top>\n<td>PubMed:</td>\n<td>", 
                         aafGetHTML(aafTab, i, "PubMed", linkText="X"), "</td>\n</tr>\n",
                         sep="")
          }
          if(annotKEGG) {
            kegg <- aafTab[i, "Pathway"]@table$Pathway[[1]]
            kstr <- "";
            if (length(kegg)!=0) {
              kegglink <- getURL(kegg)
              keggtext <- sapply(kegg, getText)
              kstr <- paste("<a href=\"", kegglink, "\" target=annotation>", 
                            keggtext, "</a><br>", sep="", collapse="")
            }
            if (length(kegg)!=0) {
              tmp <- paste(tmp, "<tr valign=top>\n<td>KEGG:</td>\n<td>", kstr,
                           "</td>\n</tr>\n", sep="")
            }
          }
          if(annotOMIM) {
            omenv    <- as.name(paste(chip, "OMIM", sep=""))
            if (exists(lookupProbes[i], eval(omenv))) { 
              omimlink <- get(lookupProbes[i],eval(omenv)) 
            } else { omimlink <-NA }
            if (is.na(omimlink[1])) {
              omim <- "NA";
            } else {
              omim <- paste("<a href=\"", OMIMlink.raw, omimlink, "\" target='annotation'>", 
                            omimlink, "</a>", sep="", collapse=", ")
            }
            if (omim!="NA") {
              tmp <- paste(tmp, "<tr valign=top><td>OMIM:</td>\n<td>", omim,
                           "</td>\n</tr>\n", sep="")
            }
          }
          if(annotGENECARDS) {
            GClink <- paste("<a href=\"", GClink.raw, probes[i], "\" target='annotation'>X</a>",
                            sep="")
            tmp <- paste(tmp, "<tr valign=top>\n<td>GeneCards:</td>\n<td>\n", GClink,
                         "</td>\n</tr>\n", sep="")
          }
          tmp <- paste(tmp, "</table>\n", sep="")
# add anchors for probes if available
	  if (!is.null(probes)) 
	      tmp <- paste("<a name='", probes[i], "'>", tmp, "</a>", sep="")
          htmlTable[i, j] <- tmp
        }
      }
      s <- paste(s, cpdTable(htmlTable, tbh=headers, spc=5, anchors=probes, 
                             stripes="horizontal"), sep="")
    } else { # so number of probes is 0
      s <- paste(s, "<p><h4>No upregulated genes among the selected ", maxgenes,
                 " most significantly up or down regulated genes.</h4></p>", sep="")
    }
  } else { # ASCII part
    n <- length(probes)
    if (n>0) {
      if (!is.null(addInformation)) {
        rowname <- colnames(addInformation)
        for (k in 1:length(rowname)) {
          s <- paste(s, toupper(rowname[k]), sep, sep="")
        }
      }

      if (annotPROBE)	     s <- paste(s, "PROBE", sep, sep="")
      if (annotSYMBOL)       s <- paste(s, "SYMBOL", sep, sep="")
      if (annotGDESCRIPTION) s <- paste(s, "DESCRIPTION", sep, sep="")
      if (annotLOCUS)        s <- paste(s, "CHRLOC", sep, sep="")
      if (annotGO)           s <- paste(s, "GENE ONTOLOGY", sep, sep="")
      if (annotCORG)         s <- paste(s, "CORG", sep, sep="")
      if (annotUNIGENE)      s <- paste(s, "UNIGENE", sep, sep="")
      if (annotGENBANK)      s <- paste(s, "GENBANK", sep, sep="")
      if (annotLOCUSLINK)    s <- paste(s, "LOCUSLINK", sep, sep="")
      if (annotPUBMED)       s <- paste(s, "PUBMED", sep, sep="")
      if (annotKEGG)         s <- paste(s, "KEGG", sep, sep="")
      if (annotOMIM)         s <- paste(s, "OMIM", sep, sep="")

      #del last tab sign
      s <- substr(s,0,nchar(s)-1)
      s <- paste(s, "\n", sep="")

      for (i in 1:n) {
        rs <- ""
        if (verbose) {
          cat(i, "/", n, "\n", sep="")
        }

        # additional information block
        if (!is.null(addInformation)) {
          rowname <- colnames(addInformation)
          for (k in 1:ncol(addInformation)) {
            rs <- paste(rs, substr(as.character(addInformation[probes[i],k]),0,6), sep, sep="")
          }
        }

        # probe information
        if (annotPROBE) {
          rs <- paste(rs, probes[i], sep, sep="")
        }
        # gene symbol information
        if (annotSYMBOL) {
          symbol <- symbol.text[i]
          if (length(symbol)==0 | symbol=="") {
            symbol <- "NA";
          }
          rs <- paste(rs, symbol, sep, sep="")
        }

        # annot GDESCRIPTION information
        if (annotGDESCRIPTION) {
          description <- description.text[i]
          if (length(description)==0 | description=="") {
            description <- "NA";
          }
  
          rs <- paste(rs, description, sep, sep="")
        }
  
        # annot LOCATION information
        if (annotLOCUS) {
          locus     <- locus.text[i]
          if (length(locus)==0 | locus=="") {
            locus <- "NA";
          }
          rs <- paste(rs, locus, sep, sep="")
        }
        # annot GO information
        if (annotGO) {
          goInfo <- getGOTable(probes[i],chip)
          goInfo <- gsub("<(.|\n)*?>","",goInfo,perl=T)
          goInfo <- gsub("\n"," ",goInfo)
          rs <- paste(rs, goInfo, sep, sep="")
        }
  
        # annot CORG information
        if (annotCORG) {
          p      <- as.name(paste(chip, "CORG", sep=""))
          if (exists(probes[i],eval(p))) { corgTF <- get(probes[i],eval(p)) }
          else { corgTF <- NA }
          if(!is.na(corgTF[1])) {
            corgStr <- "";
            if (is.matrix(corgTF)) {
              corgTF <- corgTF[,1]
            } else {
              corgTF <- corgTF[1]
            }
            for (k in 1:length(corgTF)) {
              if (k>1) corgStr <- paste(corgStr, ", ", sep="")
              orgStr <- paste(corgStr, corgTF[k], sep="")
            }
          } else {
            corgStr <- "NA"
          }
          rs <- paste(rs, corgStr, sep, sep="")
        }
  
        if(annotUNIGENE) {
          unigene     <- unigene.text[i]
          if (length(unigene)==0 | unigene=="") {
            unigene <- "NA";
          }
          rs <- paste(rs, unigene, sep, sep="")
        }

        if(annotGENBANK) {
          genbank     <- genbank.text[i]
          if (length(genbank)==0 | genbank=="") {
            genbank <- "NA";
          }
          rs <- paste(rs, genbank, sep, sep="")
        }

        if(annotLOCUSLINK) {
          locuslink     <- locuslink.text[i]
          if (length(locuslink)==0 | locuslink=="") {
            locuslink <- "NA";
          }
          rs <- paste(rs, locuslink, sep, sep="")
        }

        if(annotPUBMED) {
          pubmed     <- pubmed.text[i]
          if (pubmed=="") {
            pubmed <- "NA";
          } else {
            pstr <-""
            for (k in 1:length(pubmed)) {
              if (k>1) pstr <- paste(pstr, ", ", sep="")
              pstr <- paste(pstr, pubmed[k], sep="")
            }
          }
          if (pubmed!="NA") {
            rs <- paste(rs, pstr, sep, ssep="")
          } else {
            rs <- paste(rs, "NA", sep, sep="")
          }
        }

        if(annotKEGG) {
          kegg     <- aafPathway(probes[i],chip)[[1]]
          kstr <- "";
          if (length(kegg)!=0) {
            for (k in 1:length(kegg)) {
              if (k>1) kstr <- paste(kstr, ", ", sep="")
              kstr <- paste(kstr, kegg[[k]]@name, sep="")
            }
          }
          if (length(kegg)!=0) {
          rs <- paste(rs, kstr, sep, sep="")
          } else {
            rs <- paste(rs, "NA", sep, sep="")
          }
        }

        if(annotOMIM) {
          omenv    <- as.name(paste(chip, "OMIM", sep=""))
          if (exists(probes[i],eval(omenv))) { omimlink <- get(probes[i],eval(omenv)) }
          else { omimlink <- NA }

          if (is.na(omimlink[1])) {
            omim <- "NA";
          } else {
            omim <- ""
            for (k in length(omimlink)) {
              if (k>1) {omim <- paste(omim, ", ", sep="")}
              omim <- paste(omim, omimlink[k], sep="")
            }
          }
          rs <- paste(rs, omim, sep, sep="")
        }

        #del last tab sign
        rs <- substr(rs,0,nchar(rs)-1)
        rs <- paste(rs, "\n", sep="")
        s <- paste(s, rs, sep="")
      }
    }
  }

  if (length(grep("geneLister", sys.calls())) == 1) {
    # only add data package version, if geneLister didn't call itself
    s <- paste(s, logDataPackageVersion(chip), sep="")
  }
  return(s);
}

