# Function to compute distance of samples based on Isis-Result:
# weight differences by respective DLD-scores
weightedHammingDist <- function(ires,score.weight=3){
  # ires: ISIS-result
  binaryMatrix <- ires[,-ncol(ires)]
  splitScores <- ires[,ncol(ires)]
  whd <- function(x,y,weights){
    return(as.numeric(t(abs(x-y))%*%(weights^score.weight))) # put more weight on higher scores, exponent is arbitrary!
  }
  matwhd <- function(thiscol,weights){
    return(apply(binaryMatrix,2,whd,y=thiscol,weights=weights))
  }
  distMatrix <- apply(binaryMatrix,2,matwhd,weights=splitScores)
  return(as.dist(distMatrix))
}

# Function to plot Isis-Result:
# samples clustered using weighted Hamming distance
plotIsis <- function(ires,data,score.weight=3, horizontal=TRUE, main=NULL){
  # ires: ISIS-result
  # data: data ISIS was run on
  if (ncol(data)!=(ncol(ires)-1))
    stop("This data was probably not used for ISIS! Check arguments!\n")
  samples <- colnames(data)  
  rm(data)
  binaryMatrix <- ires[,-ncol(ires)]
  colnames(binaryMatrix) <- samples
  splitScores <- ires[,ncol(ires)]
  cat("Clustering...")
  sampledist <- weightedHammingDist(ires,score.weight)
  sampleclust <- hclust(sampledist,method="average")
  cat("Plotting...\n")
  imageMatrix <- function(mat){
    reversed <- c()
    for (i in 1:nrow(mat)){
      reversed <- rbind(mat[i,],reversed)
    }#for
    return(t(reversed))
  } #imageMatrix
#  x11(height=12,width=12)
  if (is.null(main)) tSpace <- 0 else tSpace <- 2
  nsamples <- length(samples)
  nsplits <- length(splitScores)
  if (horizontal) {
    par(ann=FALSE,mar=c(8, 0.5, 1+tSpace, 7),las=2)
    image(x=1:ncol(binaryMatrix),y=1:nrow(binaryMatrix),
          imageMatrix(binaryMatrix[,sampleclust$order]),
          col=c("steelblue","springgreen"), axes=FALSE)
    axis(labels=substr(samples[sampleclust$order],1,13), side=1, at=1:nsamples)
    axis(labels=as.character(round(rev(splitScores),digits=4)), at=1:nsplits, side=4)
    mtext(side=4,at=nsplits+1,text="DLD-Score", cex=1,font=2,las=1,line=0.3)
  } else { # samples on y axis
    par(ann=FALSE,mar=c(3, 5, 3+tSpace, 2),las=1)
    image(x=1:nrow(binaryMatrix),y=1:ncol(binaryMatrix), binaryMatrix[,sampleclust$order],
          col=c("steelblue","springgreen"), axes=FALSE)
    axis(paste("Split", 1:nsplits), at=1:nsplits, side=3)
    axis(substr(samples[sampleclust$order],1,13),at=1:nsamples, side=2, cex.axis=0.7)
    axis(as.character(round(splitScores,digits=4)), at=1:nsplits, side=1)
  }
  if (!is.null(main)) title(main=main)
  makegrid<-function(groups){
    for(i in 1:(ncol(groups)-1)){
      abline(v=i+0.5,col="white",lwd=0.1)
    }#for
    for(i in 1:(nrow(groups)-1)){
      abline(h=i+0.5,col="white",lwd=0.1)
    }#for
  }#makegrid
  makegrid(binaryMatrix)
  box()
} #plotIsis


# Function to approximate null distribution of DLD-scores
randomDLDscores <- function(data,nsplits=100,ngenes=75,ignore.genes=5){
  require(isis)
  cat(paste("Generating",nsplits,"random splits...\n")) 
  nsamples <- ncol(data)
  randomSplits <- matrix(nrow=nsplits,ncol=nsamples)
  for (i in 1:nsplits){
    oneRandomSplit <- numeric(nsamples)
    while ((sum(oneRandomSplit)>(nsamples-2))|(sum(oneRandomSplit)<2))
      oneRandomSplit <- round(runif(nsamples))
    randomSplits[i,] <- oneRandomSplit
  }#for
  rownames(randomSplits) <- NULL
  #return(randomSplits)
  cat("Calculating DLD-scores for random splits...\n")
  randScores <- tscore(data,randomSplits,p=ngenes,p.offs=ignore.genes)
  return(randScores)
} #randomDLDscores


# Function to evaluate splits by looking for differentially expressed genes
#  between the two groups, use 'twilight' for that
evalSplitsTwilight <- function(ires,data,B=1000){
  # ires: ISIS-Result
  # data: expression data used for ISIS
  # B: number of permutations in twilight
  require(twilight)
  binaryMat <- ires[,-ncol(ires)]
  splits.p0 <- c()
  splits.ncandidates <- c()
  splits.minq <- c()
  nsplits <- nrow(ires)
  for (i in 1:nsplits){
    cat(paste("Evaluating split",i,"of",nsplits,"...\n"))
    twresult <- twilight.pval(data,binaryMat[i,],method="z",B=B)
    splits.p0 <- c(splits.p0,twresult$pi0)
    ncands <- sum(twresult$result$candidate)
    splits.ncandidates <- c(splits.ncandidates,ncands)
    splits.minq <- c(splits.minq, min(twresult$result$qvalue))
    if (ncands > 0)
      cat(paste("\nFound",ncands,"candidates in split",i,"!\n"))
  } # for
  names(splits.p0)=names(splits.ncandidates)=names(splits.minq) <- paste("split",1:nsplits,sep="")
  return(list(p0 = splits.p0, ncandidates = splits.ncandidates, minq = splits.minq))
} # evalSplitsTwilight

# Function, to mark redundant splits:
removeRedundantSplits <- function(ires){
  #ires: ISIS-result
  areSplitsEqual <- function(s1,s2) {
    #s1,s2: two ISIS-splits (binary vectors)
    return( all(s1==s2) | all(s1==(s2-1)) )
  } # areSplitsEqual
  binMatrix <- ires[,-ncol(ires)]
  equalPairs <- c()
  cat("Searching for equal splits...\n")
  for (i in 1:(nrow(binMatrix)-1)){
    for (j in (i+1):nrow(binMatrix)){
      if (areSplitsEqual(binMatrix[i,],binMatrix[j,])){
        #cat(paste("Splits",i,"and",j,"are equal!\n"))
        equalPairs <- rbind(equalPairs,c(i,j))
      } #then
    } #for j
  }# for i
  #return(equalPairs)
  if (is.null(equalPairs)){
    cat("None of the splits are equal!\n")
    return(ires)
  } # if (is.null(equalPairs))
  cat("Removing",nrow(equalPairs),"redundant splits...\n")
  rownames(ires)[equalPairs[,1]] <- paste(rownames(ires)[equalPairs[,1]],rownames(ires)[equalPairs[,2]],sep=".")
  newires <- ires[-equalPairs[,2],,drop=FALSE]
  return(newires)
} # removeRedundantSplits
