runPreprocessing   <- function(BASEPATH,
                               COVDESC = paste(BASEPATH, "celfiles.lst", sep="/"),
                               TMPPATH ="/scratch/local",
                               EXPRSET = paste(BASEPATH, "expression_set.rdat", sep="/"),
                               METHOD="rma",
                               plot.preprocessing=TRUE,
                               run.simpleaffy=TRUE,
                               cleanup.plmsets = TRUE) {
  require(affy)
  require(affyPLM)
  require(vsn)
  require(docval)
  require(simpleaffy)

  cat("Preprocessing", BASEPATH, "\n")
  pp.object <- try(preprocessProject(makeAbsolute(COVDESC),
                                     makeAbsolute(BASEPATH),
                                     makeAbsolute(TMPPATH),
                                     makeAbsolute(EXPRSET),
                                     run.simpleaffy,
                                     METHOD))
  if (inherits(pp.object,"try-error") | !is.null(pp.object$e.message)) {
    warning("Something went wrong while preprocessing data\n")
    return(FALSE)
  }

  if (plot.preprocessing) {
    tmp <- try(plotProject(pp.object))
    if (inherits(tmp,"try-error")) {
      warning("Something went wrong while creating diagnostic plots\n")
      return(FALSE)
    }
  }
  
  if (run.simpleaffy) {
    tmp <- try(simpleaffyAnalysis(pp.object$TMPPATH,pp.object$SAYPATH))
    if (inherits(tmp,"try-error")) {
      warning("Something went wrong while running simpleaffy\n")
      cat("Please remove directory", pp.object$TMPPATH, "manually\n")
      return(FALSE)
    }
    file.remove(paste(pp.object$TMPPATH,"rawdata.rdat",sep="/"))
    file.remove(pp.object$TMPPATH)
    if (cleanup.plmsets) {
      file.remove(paste(pp.object$NRMPATH, "abo.rdat", sep="/"))
      file.remove(paste(pp.object$SUMPATH, "plm.rdat", sep="/"))
    }
  }
  return(invisible(TRUE))
}

makeAbsolute <- function(fname) {
  if (substr(fname, 1, 1) == "/") return(fname)
  if (substr(fname, 1, 1) == "~") {
    tmp.dir <- getwd()
    setwd("~")
    res <- getwd()
    setwd(tmp.dir)
    return(sub("~", res, fname))
  }
  return(paste(getwd(), fname, sep="/"))
}

checkConsistency   <- function(COVDESC,
                               BASEPATH,
                               TMPPATH,
                               run.simpleaffy) {
  e.message <- NULL
  
  cat(paste("Checking",COVDESC,"\n"))
  # - if COVDESC does not exist : exit
  if(!file.exists(COVDESC)){
    e.message <-
      paste(e.message,
            "\n--- ERROR: file ",COVDESC," does not exist\n",sep="")
  }
  cat(paste("Checking",BASEPATH,"\n"))
  # - if BASEPATH does not exist : exit
  if(!file.exists(paste(BASEPATH,".",sep="/"))){
    e.message <-
      paste(e.message,
            "\n--- ERROR: directory ",BASEPATH," does not exist\n",sep="")
  }
  
  # - set the path to write in normalized data
  tmpTime   <- Sys.time();class(tmpTime)<-NULL
  NRMPATH   <- paste(BASEPATH,"/normalized",sep="")
  SUMPATH   <- paste(BASEPATH,"/summarized",sep="")
  if(run.simpleaffy) {
    SAYPATH   <- paste(BASEPATH,"/simpleaffy",   sep="")
    TMPPATH   <- paste(TMPPATH,"/cpdiagTmpPrePro",tmpTime,sep="")
  } else {
    SAYPATH   <- NULL
    TMPPATH   <- NULL
  }
  cat(paste("Checking",NRMPATH,"\n"))
  # - if NORMPATH does not exist: exit
  if(!file.exists(paste(NRMPATH,".",sep="/"))){
    e.message <-
      paste(e.message,
            "\n--- ERROR: directory ",NRMPATH," does not exist\n",sep="")
  }
  cat(paste("Checking",SUMPATH,"\n"))
  # - if SUMPATH does not exist: exit
  if(!file.exists(paste(SUMPATH,".",sep="/"))){
    e.message <-
      paste(e.message,
            "\n--- ERROR: directory ",SUMPATH," does not exist\n",sep="")
  }
  if (run.simpleaffy) {
    cat(paste("Checking",SAYPATH,"\n"))
    # - if SAYPATH does not exist: exit
    if(!file.exists(paste(SAYPATH,".",sep="/"))){
      e.message <-
        paste(e.message,
              "\n--- ERROR: directory ",SAYPATH," does not exist\n",
              sep="")
    }
    cat(paste("Checking",TMPPATH,"\n"))
    # - if TMPPATH does already exist: exit
    if(file.exists(paste(TMPPATH,".",sep="/"))){
      e.message <-
        paste(e.message,
              "\n--- ERROR: directory ",TMPPATH," does already exist\n",
              sep="")
    }
  }
  return(list
         (e.message=e.message,
          NRMPATH=NRMPATH,
          SUMPATH=SUMPATH,
          SAYPATH=SAYPATH,
          TMPPATH=TMPPATH)
         )
}

preprocessProject  <- function(COVDESC,
                               BASEPATH,
                               TMPPATH,
                               EXPRSET,
                               run.simpleaffy,
                               METHOD) {

  # control random number generator
  my.seed <- ceiling(runif(1, 0, 2^20))
  set.seed(my.seed)

  paths <- checkConsistency(COVDESC,
                            BASEPATH,
                            TMPPATH,
                            run.simpleaffy)
  paths$EXPRSET <- EXPRSET

  if(!is.null(paths$e.message)) {
    cat("\nThere are inconsistencies in your directory structure\n")
    cat(paths$e.message)
    return(paths)
  }

  NRMPATH <- paths$NRMPATH
  SUMPATH <- paths$SUMPATH
  SAYPATH <- paths$SAYPATH
  TMPPATH <- paths$TMPPATH

  ## CALCULATIONS
  ##-------------
  celfiles <- getCelNames(COVDESC, 2)

  cat(date(), "Reading Data ...\n")
  abo      <- ReadAffy(filenames=celfiles, verbose=TRUE)
  sampleNames(abo) <- getCelNames(COVDESC, 1)

  ## use documentation by value for preprocessing
  cat(date(), "Normalizing ...\n")
  normalize.AffyBatch.vsn.val <- get("normalize.AffyBatch.vsn.val", asNamespace("docval"))
  abo.normalized <- normalize.AffyBatch.vsn.val(abo, subsample=20000, 
                                                niter=4, lts.quantile=0.5)

  vsnParams <- abo.normalized@description@preprocessing$vsnParams[1,,]
  rownames(vsnParams) <- sampleNames(abo.normalized)
  colnames(vsnParams) <- c("Offset", "Coefficient")
  abo.normalized@description@preprocessing$vsnParams <- vsnParams

  # save normalized AffyBatch
  filename <- paste(NRMPATH,"/abo.rdat",sep="")
  cat(date(), "Saving ", filename, "...\n")
  save(abo.normalized,file=filename, compress=TRUE)
  write.table(vsnParams, file=file.path(NRMPATH,"vsnParams.tab"), sep="\t")

  ## - create Temp-directory for storing AffyBatch
  cat(date(), "Storing temporary data in",TMPPATH, "...\n")
  dir.create(TMPPATH)
  filename <- paste(TMPPATH,"/rawdata.rdat",sep="")
  save(abo,file=filename, compress=TRUE)

  ## free memory by deleting rawdata
  rm(abo)

  ## - use affyPLM to summerize
  if (METHOD=="rma") {
    cat(date(), "Fitting probe-set models (median polish) ...\n")
    if(.Platform$OS.type!="windows") sink("/dev/null")
    plmset   <- rmaPLM(abo.normalized,
                       background=FALSE,
                       normalize =FALSE,
                       output.param=list(weights=TRUE,pseudo.SE=TRUE))
    if(.Platform$OS.type!="windows") sink()
  } else {
    cat(date(), "Fitting probe-set models (linear models) ...\n")
    plmset   <- fitPLM(abo.normalized,
                       background = FALSE,
                       normalize = FALSE)
  }

  filename <- paste(SUMPATH,"/plm.rdat",sep="")
  cat(date(), "Saving ", filename, "...\n")
  save(plmset,file=filename, compress=TRUE)

  ## - write chip effect estimates as expression matrix with vsm documentation into
  ##   an expression set with some description
  expr.set <- abo.normalized
  exprs(expr.set) <- coefs(plmset)
  expr.set@description@preprocessing$scale <- "log2"
  expr.set@description@preprocessing$cdp.seed <- my.seed
  expr.set@description@preprocessing$rmaPLM.settings <-
                                                   plmset@model.description$modelsettings
  expr.set@description@preprocessing$val$probe.effects <- coefs.probe(plmset)

  cat(date(), "Saving ", EXPRSET, "...\n")
  save(expr.set, file=EXPRSET, compress=TRUE)

  return(paths)
}

addonPreprocessing   <- function(BASEPATH,
                                 COVDESC = file.path(BASEPATH, "celfiles_addon.lst"),
                                 EXPRSET = file.path(BASEPATH,"expression_set.rdat"),
                                 plot.preprocessing=TRUE) {
  require(affy)
  require(affyPLM)
  require(vsn)
  require(docval)

  # make paths absolute
  BASEPATH <- makeAbsolute(BASEPATH)
  COVDESC  <- makeAbsolute(COVDESC)
  EXPRSET  <- makeAbsolute(EXPRSET)
                                 
  # control random number generator
  my.seed <- ceiling(runif(1, 0, 2^20))
  set.seed(my.seed)

  paths <- checkConsistency(COVDESC, BASEPATH, NULL, FALSE)
  if(!is.null(paths$e.message)) {
    cat("\nThere are inconsistencies in your directory structure\n")
    cat(paths$e.message)
    return(FALSE)
  }
  paths$EXPRSET <- EXPRSET
  COVDESC <- file.path(BASEPATH, "celfiles_addon.lst")

  NRMPATH <- paths$NRMPATH
  SUMPATH <- paths$SUMPATH

  ## CALCULATIONS
  ##-------------
  load(EXPRSET)
  scl <- preproc(description(expr.set))$val
  vsnParams <- expr.set@description@preprocessing$vsnParams

  celfiles <- getCelNames(COVDESC, 2)
  sampleNames <- getCelNames(COVDESC, 1)
  CURRENT_PATH <- getwd();
  setwd(BASEPATH)
  load(file.path(SUMPATH, "med.sum.rda"))
  load(file.path(NRMPATH, "med.nrm.rda"))
  for (i in 1:length(celfiles)) {
    cat(date(), "Addon processing", sampleNames[i], "...\n")

    # normalize
    abo <- ReadAffy(filenames=celfiles[i], verbose=FALSE)
    abo <- normalize.vsn.add(abo, scl$vsn.scale)
    if (sampleNames[i] %in% rownames(vsnParams)) {
      vsnParams[sampleNames[i],] <- abo@description@preprocessing$vsnParams
    } else {
      vsnParams <- rbind(vsnParams, abo@description@preprocessing$vsnParams)
      rownames(vsnParams)[nrow(vsnParams)] <- sampleNames[i]
    }

    # summarize
    inds.all <- indexProbes(abo, which="pm")
    if(any( names(inds.all) != names(scl$probe.effects))) 
      stop("names of the probe effects do not match the abo \n\n")
    inds <- unlist(inds.all)

    ##=== subtract the probe effects 
    exprs(abo)        <- log2(exprs(abo))   # rma on log2 scale
    abo.norm.pm       <- exprs(abo)[inds,]
    exprs(abo)[inds,] <- exprs(abo)[inds,]-unlist(scl$probe.effects)
    ##=== calculate probe-set specific median as chip effect and residuals
    chipfu            <- function(x) median(exprs(abo)[x,])
    resfu             <- function(x) exprs(abo)[x,] - median(exprs(abo)[x,])
    chip.effects      <- unlist(lapply(inds.all,chipfu))
    residuals         <- unlist(lapply(inds.all,resfu ))

    # add to expr.set
    if (sampleNames[i] %in% sampleNames(expr.set)) {
      exprs(expr.set)[,sampleNames[i]] <- chip.effects # replace existing expr. values
    } else { # new sample, to be added
      exprs(expr.set) <- cbind(exprs(expr.set), chip.effects)
      pData(expr.set)["chip.effects",] <- rep(NA, ncol(pData(expr.set)))
      sampleNames(expr.set)[length(sampleNames(expr.set))] <- sampleNames[i]
    }

    # draw chip-specific plots
    if (plot.preprocessing) {
      MvsA.cpd(abo.norm.pm,  med.nrm,call="normalization", 
               filename=file.path(NRMPATH, sampleNames[i]))
      MvsA.cpd(chip.effects, med.sum,call="summary", 
               filename=file.path(SUMPATH, sampleNames[i]))
      wimage.resids(sampleNames[i], residuals, abo, inds, 
                    filename=file.path(SUMPATH, sampleNames[i]))
    }
  }  
  setwd(CURRENT_PATH)

  # update expr.set
  expr.set@description@preprocessing$vsnParams <- vsnParams
  write.table(vsnParams, file=file.path(NRMPATH, "vsnParams.tab"), sep="\t")
  if (require(pipelineTools)) {
    postPPL <- parsePPPLConfigFile(file.path(BASEPATH, "project.config"))
    tmp <- getPhenoData(sampleNames(expr.set), postPPL@pheno_data_file, 
                        postPPL@pheno_desc_file)
    if (is.null(tmp$phenoDesc)) tmp$phenoDesc <- colnames(tmp$phenoData)
    phenoData(expr.set) <- new("phenoData", pData=tmp$phenoData, varLabels=tmp$phenoDesc)
  } else {
    warning("Package pipelineTools not found, phenoData not updated")
  }
  expr.set@description@preprocessing$cdp.addon.seed <- my.seed

  cat(date(), "Saving ", EXPRSET, "...\n")
  save(expr.set, file=EXPRSET, compress=TRUE)

  if (plot.preprocessing) {
    expr.mat <- exprs(expr.set)
    setwd(SUMPATH)
    boxplt.cpd(expr.mat,
               filename  = "boxplot.png",
               call      = "summary",
               width     = 10,
               pointsize = 8)
    boxplt.cpd(expr.mat,
               filename  = "boxplot.pdf",
               call      = "summary",
               width     = 10,
               pointsize = 8,
               type      = "pdfwrite")
  }
  return(invisible(TRUE))
}

plotProject        <- function(preproDataPath = list()) {

  CURRENT_PATH <- getwd();

  if(!is.null(preproDataPath$e.messages))
    return(FALSE)

  SUMPATH <- preproDataPath$SUMPATH
  NRMPATH <- preproDataPath$NRMPATH
  TMPPATH <- preproDataPath$TMPPATH
  EXPRSET <- preproDataPath$EXPRSET

  cat(date(), "Loading data for plotting ...\n")
  load(EXPRSET)
  expr.mat <- exprs(expr.set)
  load(paste(NRMPATH,"abo.rdat",sep="/"))
  load(paste(SUMPATH,"plm.rdat",sep="/"))
  load(paste(TMPPATH,"/rawdata.rdat",sep=""))

  ## ------------------
  ## - do some plotting
  ## ------------------
  cat(date(), "Boxplot of summarized probesets ...\n")
  width.pass  = 10       # inches
  pointsize.pass <- 8
  setwd(SUMPATH)
  filename.pass <- "boxplot.png"
  boxplt.cpd(expr.mat,
             filename  = filename.pass,
             call      = "summary",
             width     = width.pass,
             pointsize = pointsize.pass)
  filename.pass <- "boxplot.pdf"
  boxplt.cpd(expr.mat,
             filename  = filename.pass,
             call      = "summary",
             width     = width.pass,
             pointsize = pointsize.pass,
             type      = "pdfwrite")

  ## - produce boxplot on probe level
  cat(date(), "Boxplot at probe level ...\n")
  abo.norm.pm <- log2(exprs(abo.normalized)[unlist(indexProbes(abo.normalized,
                                                            "pm")),])
  width.pass  = 10       # inches
  pointsize.pass <- 8
  setwd(NRMPATH)
  filename.pass <- "boxplot.png"
  boxplt.cpd(abo.norm.pm,
             filename  = filename.pass,
             call      = "normalization",
             width     = width.pass,
             pointsize = pointsize.pass)
  filename.pass <- "boxplot.pdf"
  boxplt.cpd(abo.norm.pm,
             filename  = filename.pass,
             call      = "normalization",
             width     = width.pass,
             pointsize = pointsize.pass,
             type      = "pdfwrite")
  ## - produce boxplot for raw data
  cat(date(), "Boxplot raw data ...\n")
  abo.pm <- log(exprs(abo)[unlist(indexProbes(abo, "pm")),])
  width.pass  = 10       # inches
  pointsize.pass <- 8
  setwd(NRMPATH)
  filename.pass <- "raw.boxplot.png"
  boxplt.cpd(abo.pm,
             filename  = filename.pass,
             call      = "raw data",
             width     = width.pass,
             pointsize = pointsize.pass)
  filename.pass <- "raw.boxplot.pdf"
  boxplt.cpd(abo.pm,
             filename  = filename.pass,
             call      = "raw data",
             width     = width.pass,
             pointsize = pointsize.pass,
             type      = "pdfwrite")

  ## - produce mva-plots
  cat(date(), "Probe-set level rank-residual-Plots ...\n")
  setwd(SUMPATH)
  med.sum <- apply(expr.mat,1,median)
  save(med.sum, file="med.sum.rda", compress=TRUE)
  for (i in 1:ncol(expr.mat)) {
    cat(".")
    filename=colnames(expr.mat)[i]
    MvsA.cpd(expr.mat[,i],med.sum,call="summary",filename=filename)
#    MvsA.cpd(expr.mat[,i],med.sum,call="summary",filename=filename, type="pdfwrite")
  }

  cat("\n", date(), "Probe level rank-residual-Plots ...\n")
  med.nrm <- apply(abo.norm.pm,1,median)
  setwd(NRMPATH)
  save(med.nrm, file="med.nrm.rda", compress=TRUE)
  for (i in 1:ncol(abo.norm.pm)) {
    cat(".")
    filename=colnames(abo.norm.pm)[i]
    MvsA.cpd(abo.norm.pm[,i],med.nrm,call="normalization",filename=filename)
#    MvsA.cpd(abo.norm.pm[,i],med.nrm,call="normalization",filename=filename,
#             type="pdfwrite")
  }

  ## - produce residual-plots from affyPLM
  cat("\n", date(), "Residuals vs chip coordinate plots ...\n")
  setwd(SUMPATH)
  for (i in 1:ncol(expr.mat)) {
    cat(".")
    filename=colnames(expr.mat)[i]
    wimage.cpd(plmset,which=i,filename=filename)
  }

  ## - produce se-boxplot from affyPLM
  cat("\n", date(), "Probe residuals boxplot ...\n")
  width.pass  = 10       # inches
  pointsize.pass <- 8
  setwd(SUMPATH)
  filename.pass <- "boxplot_plm.png"
  boxplt.plmset.cpd(plmset,
                    filename  = filename.pass,
                    width     = width.pass,
                    pointsize = pointsize.pass)
  filename.pass <- "boxplot_plm.pdf"
  boxplt.plmset.cpd(plmset,
                    filename  = filename.pass,
                    width     = width.pass,
                    pointsize = pointsize.pass,
                    type      = "pdfwrite")

  setwd(CURRENT_PATH)
  return(TRUE)
}

simpleaffyAnalysis <- function(TMPPATH,
                               SAYPATH) {

  CURRENT_PATH = getwd()

  cat(date(), "Running simpleaffy analysis ...\n")
  library(simpleaffy)

  filename <- paste(TMPPATH,"/rawdata.rdat",sep="")
  load(filename)

  setwd(SAYPATH)

  abo.mas5 <- call.exprs(abo,"mas5")
  qc <- qc(abo,abo.mas5)
  bitmap(file="qc.png", height=1.6+length(sampleNames(abo))/8, width=10, 
         pointsize=8, res=144)
  plot(qc,label=sampleNames(abo),bg.thresh=25)
  dev.off()
  bitmap(file="qc.pdf",  pointsize=8, height=1.6+length(sampleNames(abo))/6, width=10, type="pdfwrite")
  plot(qc,label=sampleNames(abo),bg.thresh=25)
  dev.off()

  qc.table <- cbind(avbg(qc),
                    sfs(qc),
                    percent.present(qc),
                    qcProbes(qc))
  colnames(qc.table) <- c("average_background",
                          "scale_factors",
                          "percent_present",
                          "actin.3", "actin.M", "actin.5",
                          "gapdh.3", "gapdh.M", "gapdh.5")

  ## - save results from mas5-quality check
  write.table(qc.table,"qc.tab")
  save(qc,file="qc.rdat", compress=TRUE)

  setwd(CURRENT_PATH)

}

"MvsA.cpd" <-
function(a.x, a.y, call="", filename=NULL, sample.size = 10000,
         win.size=1000, win.step=trunc(length(a.x)/50), ...){
  require(hexbin) || stop("Package hexbin unavailable")

  ## - calculate m and a
  m <- a.x-a.y
  a <- rank(a.x+a.y)
  m.o <- m[a]

  ## sampling
  if (!is.null(sample.size) && (sample.size < length(m))) {
    s <- sample(seq(length(m)), sample.size)
    m <- m[s]
    a <- a[s]
  }

  ## compute running variance
  v.x <- seq(1, length(m.o)-win.size, win.step)
  v.y <- NULL
  for (i in v.x) v.y <- c(v.y, 3*mad(m.o[i:(i+win.size)]))
  v.x <- v.x + win.size/2

  ## - set terminal / output file depending where call comes from and according to type
  filetype <- "png"
  elipse <- list(...)
  if (!is.null(elipse$type)) {
    filetype <- elipse$type
    if (filetype == "pdfwrite") filetype = "pdf"
  }
  if(call == "normalization"){
    if(!is.null(filename)){
      filename.pass <- paste(filename,"_mva_nrm.",filetype,sep="")
      bitmap(file=paste('"',filename.pass,'"',sep=""), ...)
      par(mai=c(.9,.9,.7,.7))}
   }
  
  if(call == "summary"){
    if(!is.null(filename)){
      filename.pass <- paste(filename,"_mva_sum.",filetype,sep="")
      bitmap(file=paste('"',filename.pass,'"',sep=""), ...)
      par(mai=c(.9,.9,.7,.7))}
  }

  ## - plot mva
  select <- (m > -2) & (m < 2)
  xbins <- 100
  hbin  = hexbin(c(0,0,a[select]),c(-2,2,m[select]),xbins = xbins) # hack to force yscale
  pars = plot(hbin, main="MvsA", xlab="rank(h_chip + h_median)", 
              ylab=expression(Delta*h))
  pushHexport(pars$plot.vp)
  fit <- lowess(a,m, f=.5, iter=2)
  grid.lines(fit$x, fit$y, default.units="native", gp=gpar(col="red"))
  grid.lines(v.x, v.y, default.units="native", gp=gpar(col="blue"))
  grid.lines(c(min(a), max(a)), c(0,0), default.units="native", gp=gpar(col="black"))
  popViewport()

  ## - close device
  if(!is.null(filename))
    dev.off()
  return(invisible(pars$plot.vp))
}

"boxplt.plmset.cpd" <-
function(plmset, filename=NULL, pointsize=14, height = 1.6+ncol(coefs(plmset))*pointsize/48, ...){
  fr <- plmset

  ## - choose device
  if(!is.null(filename)) 
    bitmap(file=paste('"',filename,'"',sep=""), height=height, pointsize=pointsize, ...)
  par(las=1, mai=c(.9,2.0,.7,.7))

  ## - plotting
  boxplot(fr, horizontal=TRUE, pch=".", xlab="probe residuals")

  ## - close device
  if (!is.null(filename)) dev.off()
}

"boxplt.cpd" <-
function(fr, filename=NULL, pointsize=14, height = 1.6+ncol(fr)*pointsize/48, call="", ...){
  fr <- data.frame(fr)

  ## - choose device
  if(!is.null(filename)) 
    bitmap(file=paste('"',filename,'"',sep=""), height=height, pointsize=pointsize, ...)
  par(las=1, mai=c(.9,2.0,.7,.7))

  ## - plotting
  boxplot(fr, horizontal=TRUE, pch=".", xlab="expression levels")

  ## - close device
  if (!is.null(filename)) dev.off()
}

"wimage.cpd" <-
function(plmset,which,filename=NULL,...) {
  if(!is.null(filename)){
      filename.pass <- paste(filename,"_resids.png",sep="")
      bitmap(file=paste('"',filename.pass,'"',sep=""),...)
      #par(mai=c(.9,.9,.7,.7))}
    }
  image(plmset,which=which,type="resids",add.legend=TRUE)
  if (!is.null(filename)) dev.off()
}

wimage.resids <- function(sampleName, residuals, abo, pm.index, filename=NULLm, ...) {
  require(affyPLM)

  col.resids <- pseudoPalette(low="blue",high="red",mid="white")

  rows <- abo@nrow
  cols <- abo@ncol
  pm.x.locs <- pm.index%%rows
  pm.x.locs[pm.x.locs == 0] <- rows
  pm.y.locs <- pm.index%/%rows + 1
  xycoor <- matrix(cbind(pm.x.locs,pm.y.locs),ncol=2)
  xycoor2 <- matrix(cbind(pm.x.locs,pm.y.locs+1),ncol=2)

  resid.range <- range(residuals)
  residsmatrix <- matrix(nrow=rows,ncol=cols)
  residsmatrix[xycoor] <- residuals
  residsmatrix[xycoor2]<- residuals
  residsmatrix<- as.matrix(rev(as.data.frame(residsmatrix)))

  if(!is.null(filename)){
    filename.pass <- paste(filename,"_resids.png",sep="")
    bitmap(file=paste('"',filename.pass,'"',sep=""),...)
  }

  layout(matrix(c(1, 2), 1, 2), width = c(9, 1))
  par(mar = c(4, 4, 5, 3))
  residsmatrix <- sign(residsmatrix)*log2(abs(residsmatrix)+1)
  main.cur=sampleName
  image(residsmatrix, col=col.resids, xaxt='n', yaxt='n', main=main.cur, 
        zlim=c(-max(log2(abs(resid.range)+1)), max(log2(abs(resid.range)+1))))
  par(mar = c(4, 0, 5, 3))
  pseudoColorBar(seq(-max(log2(abs(resid.range)+1)), max(log2(abs(resid.range)+1)),0.1), 
                 horizontal = FALSE, col = col.resids, main = "",log.ticks=TRUE)
  layout(1)
  par(mar = c(5, 4, 4, 2) + 0.1)
                  
  if (!is.null(filename)) dev.off()
}

getCelNames <- function(cel_file_list, which=1) {
  covdesc <- readLines(cel_file_list)
  select  <- -c(grep("^#", covdesc),       # remove comment lines
                grep("^[ \t]*$", covdesc)) # remove empty lines
  if (length(select) > 0) covdesc <- covdesc[select]
  covdesc <- strsplit(covdesc, "\t")
  return(sapply(covdesc, function(el) el[which]))
}

