runCDNAPreprocessing <- function(BASEPATH,
                 COVDESC=paste(BASEPATH, "celfiles.lst", sep="/"),
		 EXPRSET=paste(BASEPATH, "expression_set.rdat", sep="/"),
		 TMPPATH="/scratch/local",
		 METHOD="loess",
		 CHIP="NA",
		 plot.preprocessing=TRUE) {

  require(vsn)
  require(marray)

  # try preprocessing...
  pp.object <- try(preprocessCDNAdata(makeAbsolute(COVDESC), makeAbsolute(BASEPATH), makeAbsolute(TMPPATH), makeAbsolute(EXPRSET), METHOD, CHIP))

  # if some error occured
  if (inherits(pp.object,"try-error") | !is.null(pp.object$e.message)) {
    warning("Something went wrong while preprocessing data\n")
    return(FALSE)
  }

  # plotting preprocessing
  if (plot.preprocessing) {
    tmp <- try(plotCDNAdata(pp.object))
    if (inherits(tmp,"try-error")) {
      warning("Something went wrong while creating diagnostic plots\n")
      return(FALSE)
    }
  }

  # cleaning up
  file.remove(paste(pp.object$TMPPATH,"rawdata.rdat",sep="/"))
  file.remove(pp.object$TMPPATH)

  return(TRUE)
}




# checks if pathes are correct
checkCDNAconsistency   <- function(COVDESC, BASEPATH, TMPPATH) {
  e.message <- NULL

  cat(paste("Checking",COVDESC,"\n"))
  # - if COVDESC does not exist : exit
  if(!file.exists(COVDESC)){
    e.message <-
      paste(e.message,
            "\n--- ERROR: file ",COVDESC," does not exist\n",sep="")
  }

  # path for summarized data
  SUMPATH   <- paste(BASEPATH,"/summarized",sep="")
  cat(paste("Checking",SUMPATH,"\n"))
  # - if SUMPATH does not exist: exit
  if(!file.exists(paste(SUMPATH,".",sep="/"))){
    e.message <-
      paste(e.message,
            "\n--- ERROR: directory ",SUMPATH," does not exist\n",sep="")
  }

  # path for normalized data
#  NRMPATH   <- paste(BASEPATH,"/normalized",sep="")
#  cat(paste("Checking",NRMPATH,"\n"))
  # - if NORMPATH does not exist: exit
#  if(!file.exists(paste(NRMPATH,".",sep="/"))){
#    e.message <-
#      paste(e.message,
#            "\n--- ERROR: directory ",NRMPATH," does not exist\n",sep="")
#  }

  # path for raw data
  tmpTime   <- Sys.time();class(tmpTime)<-NULL
  TMPPATH   <- paste(TMPPATH,"/cpdiagTmpPrePro",tmpTime,sep="")
  cat(paste("Checking",TMPPATH,"\n"))
  # - if TMPPATH does already exist: exit
  if(file.exists(paste(TMPPATH,".",sep="/"))){
    e.message <- paste(e.message, "\n--- ERROR: directory ",
                       TMPPATH," does already exist\n", sep="")
  }

  return(list(e.message=e.message, SUMPATH=SUMPATH,
              TMPPATH=TMPPATH, BASEPATH=BASEPATH))
}





# this variable is later used in the marray functions,
# so it is necessary that it available in the workspace, we save it
# with the layout, so that it is created when the layout is load
opt <- list()

# do the preprocessing
preprocessCDNAdata <- function(COVDESC, BASEPATH, TMPPATH, EXPRSET, METHOD, CHIP) {

  # check consisencey of paths
  paths <- checkCDNAconsistency(COVDESC, BASEPATH, TMPPATH)

  paths$EXPRSET <- EXPRSET

  TMPPATH <- paths$TMPPATH

  if(!is.null(paths$e.message)) {
    cat("\nThere are inconsistencies in your directory structure\n")
    cat(paths$e.message)
    return(paths)
  }

  # indentifers of data (xls) files -> names without .xls
  celnames <- getCelNames(COVDESC)
  # corresponding paths
  celfiles <- getCelNames(COVDESC, 2)
  # corresponding file names
  tmp <- unlist(strsplit(celfiles,"rawdata/"))
  select <- c(FALSE, TRUE)
  select <- rep(select, length(tmp)/2)
  fnames <- tmp[select]

  # check, if files named in the celfile are compressed with gz or not
  compressed <- (length(grep(".gz",celfiles))>0)

  # get unzipped filenames
  if(compressed) {
    fnames_unzip <- sub(".gz", "", fnames)
  }
  else {
    fnames_unzip <- fnames
  }

  # get data path
  data_path <- sub(paste("/",fnames[1],sep=""), "", celfiles[1])

  act_path <- getwd()

  # for some reason we have to set the current work path to the directory
  # with the raw data to read it with marray
  setwd(data_path)

  # extract files
  if(compressed) {
    cat("extracting rawdata...\n")
    for(i in 1:length(fnames)) {
      system(paste("gunzip -q -c", fnames[i], ">", fnames_unzip[i]), intern=T)
    }
  }

  # checking if userdefined layout exists
  layout_fname <- paste(CHIP,".Rda", sep="")
  if(file.exists(layout_fname)) {
    cat("using layout", CHIP, "\n")
    load(layout_fname)
    # try to load with user defined layout, this should consist of 3 variables
    # header_names (a list of the column headers), layout (a marray layout class)
    # and gnames (a list of gene names and clone ids)
    data_raw <- read.SMD(fnames=fnames_unzip, DEBUG=T, name.Gf=header_names$Gf,
                         name.Gb=header_names$Gb, name.Rf=header_names$Rf,
                         name.Rb=header_names$Rb,
                         info.id = header_names$name_clone,
                         layout=layout, gnames=gnames)
    if (inherits(data_raw,"try-error")) {
      stop("Unable to read data files using userdefined layout, please check layout files.\n")
      return(FALSE)
    }
  }
  else {
    cat("using standart layout\n")
    # try to load with standart layout (i.e. dataset Cario05 has std layout)
    data_raw <- try(read.SMD(fnames=fnames_unzip, DEBUG=FALSE))
    if (inherits(data_raw,"try-error")) {
      stop("Unable to read data files, try to build your own layout.\n")
    return(FALSE)
    }
  }

  # delete extracted files
  if(compressed) {
    for(i in 1:length(fnames_unzip)) {
      system(paste("rm -f", fnames_unzip[i]), intern=T)
    }
  }

  setwd(act_path)

  # normalization with marray
  cat("start ", METHOD, " normalization...\n")
  data_norm <- maNorm(data_raw, norm=METHOD)
  cat("...finished ", METHOD, " normalization\n")

  # remove raw data
  cat(date(), "Storing temporary data in",TMPPATH, "...\n")
  dir.create(TMPPATH)
  filename <- paste(TMPPATH,"/rawdata.rdat",sep="")
  save(data_raw,file=filename, compress=TRUE)
  rm(data_raw)

  # initialize data object, so we can delete data_norm
  data_vsn <- data_norm@maM

  # set patient (chip) name as columnnames
  colnames(data_vsn) <- celnames
  # set genenames as rownames
  colnames(data_norm@maGnames@maInfo)[1] <- "name"
  rownames(data_vsn) <- data_norm@maGnames@maInfo$name

  # impute if necessary
  if(sum(is.na(data_vsn))>0) {
    cat("start imputing...\n")
    data_vsn <- try(suppressWarnings(impute.knn(data_vsn, k=10, rowmax=0.5, colmax=0.8, maxp=3000, rng.seed=362436069)))

    if (inherits(data_vsn,"try-error")) {
      stop("Imputing failed, maybe not enough data files (patients/chips).\n")
      return(FALSE)
    }
    cat("...finished imputing\n")
   }


  # variance stabilization
  cat("start vsn...\n")
  data_vsn <- vsn(data_vsn, niter=4, lts.quantile=0.5)
  cat("...finished vsn\n")

  # vsn parameters
  vsn_params <- data_vsn@description@preprocessing$vsnParams[1,,]
  rownames(vsn_params) <- data_norm@maTargets@maLabels
  colnames(vsn_params) <- c("Offset", "Coefficient")

  data_vsn@description$vsnParams <- vsn_params
  expr.set <- data_vsn
  expr.set@description@preprocessing$scale <- "log2"

  cat(date(), "Saving ", EXPRSET, "...\n")
  save(expr.set, file=EXPRSET, compress=TRUE)

  return(paths)
}





# plotting
plotCDNAdata <- function(paths = list()) {

  EXPRSET <- paths$EXPRSET
  SUMPATH <- paths$SUMPATH
  TMPPATH <- paths$TMPPATH
  RAWPATH <- paste(paths$BASEPATH,"raw",sep="/")
  ACTPATH <- getwd()

  cat(date(), "Loading data for plotting ...\n")
  load(EXPRSET)
  load(paste(TMPPATH, "rawdata.rdat", sep="/"))

  expr.mat <- exprs(expr.set)

  intensity_g <- data_raw@maGf # refenence
  intensity_r <- data_raw@maRf # sample
  colnames(intensity_g) <- colnames(expr.mat)
  rownames(intensity_g) <- rownames(expr.mat)
  colnames(intensity_r) <- colnames(expr.mat)
  rownames(intensity_r) <- rownames(expr.mat)

  # raw data MvA plots (currently stored in the normalization folder)
  cat(date(), "Raw-set level rank-residual-Plots ...\n")
  suppressWarnings(dir.create(RAWPATH))
  setwd(RAWPATH)
  for (i in 1:ncol(intensity_g)) {
    cat(".")
    filename=colnames(intensity_g)[i]
    MvsA.cpd(log(intensity_r[,i]), log(intensity_g[,i]), call="normalization",
             filename=filename)
  }

  # summary data MvA plots (currently stored in the normalization folder)
  cat(date(), "Summarized-set level rank-residual-Plots ...\n")
  setwd(SUMPATH)
  med.sum <- apply(expr.mat, 1, median)
  for (i in 1:ncol(expr.mat)) {
    cat(".")
    filename=colnames(expr.mat)[i]
    MvsA.cpd(expr.mat[,i], med.sum, call="summary",
             filename=filename)
  }

  # plotting summarize boxplot
  cat(date(), "Boxplot of summarized probesets ...\n")
  width.pass  = 10       # inches
  pointsize.pass <- 8
  setwd(SUMPATH)
  filename.pass <- "boxplot.png"
  boxplt.cpd(expr.mat,
             filename  = filename.pass,
             call      = "summary",
             width     = width.pass,
             pointsize = pointsize.pass)
  filename.pass <- "boxplot.pdf"
  boxplt.cpd(expr.mat,
             filename  = filename.pass,
             call      = "summary",
             width     = width.pass,
             pointsize = pointsize.pass,
             type      = "pdfwrite")

  setwd(ACTPATH)

  return(TRUE)
}

